"""Increase size of mimetype field for mail attachments.

Revision ID: 46adf5114de1
Revises: 96017ac786bf
Create Date: 2017-08-17 10:57:45.173199

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '46adf5114de1'
down_revision = '96017ac786bf'


def upgrade():
    op.alter_column('attachment', 'mimetype', type_=sa.VARCHAR(100))


def downgrade():
    op.alter_column('attachment', 'mimetype', type_=sa.VARCHAR(30))



"""Add assignment todo

Revision ID: 9da4a1cc4bf1
Revises: 48303c2cf2eb
Create Date: 2023-07-30 13:32:35.422103

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '9da4a1cc4bf1'
down_revision = '48303c2cf2eb'


def upgrade():
    op.create_table(
        'assignmenttodo',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('organization_id', sa.Integer(), nullable=False),
        sa.Column('description', sa.String(), nullable=True),
        sa.Column(
            'priority', sa.Integer(), nullable=False, server_default='4'
        ),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('assignmenttodo')



"""Add ah signature

Revision ID: e2cfe9da21cf
Revises: 897a3181401e
Create Date: 2020-10-15 07:20:57.570537

"""

from alembic import op
import base64
import pkg_resources

# revision identifiers, used by Alembic.
revision = 'e2cfe9da21cf'
down_revision = '897a3181401e'


def upgrade():
    signature = base64.b64encode(
        pkg_resources.resource_stream(
            'sw.allotmentclub.signatures', 'ah.png').read()
    ).decode()
    signature = f'data:application/png;base64,{signature}'
    op.execute(f"""UPDATE public.user SET
                  signature = '{signature}'
                  WHERE username = 'ah';""")


def downgrade():
    op.execute("""UPDATE public.user SET
                  signature = ''
                  WHERE username = 'ah';""")



"""Grundgebuehr (Kraft-)Stromzaehler.

Revision ID: 63317db97d4
Revises: 444966e3e7d6
Create Date: 2014-12-25 08:20:45.348586

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '63317db97d4'
down_revision = '444966e3e7d6'


def upgrade():
    op.add_column(
        'energyprice',
        sa.Column('normal_fee', sa.Integer(), nullable=True))
    op.add_column(
        'energyprice',
        sa.Column('power_fee', sa.Integer(), nullable=True))
    op.execute("""UPDATE energyprice SET price = 3020, normal_fee = 81700,
                  power_fee = 243300 WHERE year = 2014;""")


def downgrade():
    op.drop_column('energyprice', 'power_fee')
    op.drop_column('energyprice', 'normal_fee')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from .model import Log
import datetime
import logging
import user_agents


STRIP_CHARS = '\n\r\f<:'


class UserRequestLogger(logging.LoggerAdapter):

    def __init__(self, logger):
        """Overwrite init so we do not require the dict used for extra args."""
        super(UserRequestLogger, self).__init__(logger, {})

    def __getattr__(self, name):
        """Forward unknown calls to the logger. Required for testing setup."""
        return getattr(self.logger, name)

    def process(self, msg, kwargs):
        """Enrich log msg with IP address and browser used for current request.

        Can be disabled by adding `log_request=False` as kwarg to log call.
        """
        import pyramid.threadlocal
        request = pyramid.threadlocal.get_current_request()

        if request is None or not kwargs.pop('log_request', True):
            return msg, kwargs

        ip_address = request.environ.get('HTTP_X_REAL_IP')
        if ip_address is None and hasattr(request, 'client_addr'):
            ip_address = request.client_addr
        user_agent_header = request.environ.get('HTTP_USER_AGENT')
        browser = 'no useragent sent'
        if user_agent_header is not None:
            user_agent = user_agents.parse(user_agent_header)
            family = user_agent.browser.family
            version = user_agent.browser.version_string
            browser = family
            if version:
                browser = browser + ' ' + version

        return '%s [IP:%s] [Browser:%s]' % (msg, ip_address, browser), kwargs


app_log = logging.getLogger('app')
client_log = UserRequestLogger(logging.getLogger('client'))
auth_log = UserRequestLogger(logging.getLogger('auth'))
error_log = UserRequestLogger(logging.getLogger('error'))
user_admin_log = UserRequestLogger(logging.getLogger('user admin'))
user_data_log = UserRequestLogger(logging.getLogger('user data'))


def log_with_user(log, user, *args, **kwargs):
    """Use this function to log into database, too."""
    kwargs['extra'] = {'user': user}
    return log(*args, **kwargs)


def _strip(obj):
    """Strip chars which should not be in log file due to app-12.1.6."""
    return obj.translate(None, STRIP_CHARS)


class TruncatingLogRecord(logging.LogRecord):
    """Truncating log message to a maximum value defined in environment.

    Environment key is `LOG_RECORD_MAX_LEN`.
    If not set it defaults to unlimited.

    """
    _record_max_len = None

    def getMessage(self):
        msg = super(TruncatingLogRecord, self).getMessage()
        if self._record_max_len and len(msg) > self._record_max_len:
            ellisis = '…'
            msg = msg[0:self._record_max_len] + ellisis
        return msg


class TruncatingLogAndStrippingArgsRecord(TruncatingLogRecord):

    def getMessage(self):
        self._strip_args()
        return super(TruncatingLogAndStrippingArgsRecord, self).getMessage()

    def _strip_args(self):
        if isinstance(self.args, tuple):
            self.args = tuple(x for x in self.args)
        if isinstance(self.args, dict):
            self.args = dict((k, v) for k, v in self.args.items())


class StreamHandler(logging.StreamHandler):
    """Custom stream handler which truncates log record."""

    def __init__(self, stream=None, record_max_len=0):
        super(StreamHandler, self).__init__(stream)
        self._record_max_len = int(record_max_len)

    def emit(self, record):
        # Switch the record to a TruncatingLogAndStrippingArgsRecord:
        record.__class__ = TruncatingLogAndStrippingArgsRecord
        record._record_max_len = self._record_max_len
        return super(StreamHandler, self).emit(record)


class DBHandler(logging.Handler):
    """Handler to log user related activity in the database"""

    def __init__(self, record_max_len=0):
        super(DBHandler, self).__init__()
        self._record_max_len = int(record_max_len)

    def emit(self, record):
        # Switch the record to a TruncatingLogRecord:
        record.__class__ = TruncatingLogRecord
        record._record_max_len = self._record_max_len
        user = getattr(record, 'user', None)
        if not user:
            return  # will skip all non-user centric logs like errors, SQL etc.
        message = record.getMessage()
        created = datetime.datetime.fromtimestamp(record.created)
        level = record.levelname
        name = record.name
        Log.create(msg=message, user=user, created=created, level=level,
                   name=name, organization_id=user.organization_id)



"""Add budgets to accounting

Revision ID: 53f3d16657d3
Revises: 92cec0e8663e
Create Date: 2020-12-29 14:44:13.423851

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '53f3d16657d3'
down_revision = '92cec0e8663e'

BOOKING_KINDS = {
    'Mitgliedsbeitrag': 1,
    'Energieabschlag I': 3,
    'Energieabschlag II': 4,
    'Energieabrechnung': 5,
    'Grundsteuer B': 6,
    'Abwasser': 7,
    'Aufnahmegebühr': 13,
    'Grundbesitzabgabe': 14,
    'Versicherungen': 15,
    'Verwaltung': 16,
    'Aufwandsentschädigungen': 17,
    'Instandhaltung': 18,
    'Energieanbieter': 19,
    'Müll': 20,
    'Vereinsfest': 21,
    'Pacht': 22,
    'Körperschaftssteuer': 23,
    'Arbeitsstunden': 2,
    'Mahn- & Zusatzleistungen': 24,
    'sonstige Einnahmen': 25,
    'sonstige Ausgaben': 26,
    'Satellitenanlage': 27,
}

DATA = {
    2015: {
        'Mitgliedsbeitrag': 64800000,
        'Aufnahmegebühr': 500000,
        'Mahn- & Zusatzleistungen': 0,
        'Arbeitsstunden': 15000000,
        'Körperschaftssteuer': 0,
        'sonstige Einnahmen': 600000,
        'Pacht': 6250000,
        'Verwaltung': -20000000,
        'Versicherungen': -6200000,
        'Müll': -11000000,
        'Abwasser': -700000,
        'Instandhaltung': -50000000,
        'Aufwandsentschädigungen': -8400000,
        'Grundsteuer B': -490000,
        'Grundbesitzabgabe': -650000,
        'sonstige Ausgaben': -2000000,
        'Vereinsfest': 0,
        'Satellitenanlage': -5000000,
        'Energieabrechnung': 0,
        'Energieanbieter': 0,
        'Energieabschlag I': 0,
        'Energieabschlag II': 0,
    },
    2016: {
        'Mitgliedsbeitrag': 79500000,
        'Aufnahmegebühr': 500000,
        'Mahn- & Zusatzleistungen': 0,
        'Arbeitsstunden': 10000000,
        'Körperschaftssteuer': 0,
        'sonstige Einnahmen': 0,
        'Pacht': 7000000,
        'Verwaltung': -7000000,
        'Versicherungen': -6000000,
        'Müll': -9500000,
        'Abwasser': -600000,
        'Instandhaltung': -50000000,
        'Aufwandsentschädigungen': -8600000,
        'Grundsteuer B': -490000,
        'Grundbesitzabgabe': -200000,
        'sonstige Ausgaben': -2000000,
        'Vereinsfest': 0,
        'Satellitenanlage': -5000000,
        'Energieabrechnung': 0,
        'Energieanbieter': 0,
        'Energieabschlag I': 0,
        'Energieabschlag II': 0,
    },
    2017: {
        'Mitgliedsbeitrag': 80250000,
        'Aufnahmegebühr': 1000000,
        'Mahn- & Zusatzleistungen': 0,
        'Arbeitsstunden': 13000000,
        'Körperschaftssteuer': 0,
        'sonstige Einnahmen': 0,
        'Pacht': 6203900,
        'Verwaltung': -7000000,
        'Versicherungen': -6300000,
        'Müll': -10000000,
        'Abwasser': 0,
        'Instandhaltung': -40000000,
        'Aufwandsentschädigungen': -9000000,
        'Grundsteuer B': -494100,
        'Grundbesitzabgabe': -200700,
        'sonstige Ausgaben': -3000000,
        'Vereinsfest': -20000000,
        'Satellitenanlage': -5000000,
        'Energieabrechnung': 0,
        'Energieanbieter': 0,
        'Energieabschlag I': 0,
        'Energieabschlag II': 0,
    },
    2018: {
        'Mitgliedsbeitrag': 68900000,
        'Aufnahmegebühr': 1000000,
        'Mahn- & Zusatzleistungen': 0,
        'Arbeitsstunden': 15000000,
        'Körperschaftssteuer': 50000,
        'sonstige Einnahmen': 0,
        'Pacht': 6203900,
        'Verwaltung': -7500000,
        'Versicherungen': -6000000,
        'Müll': -12000000,
        'Abwasser': 0,
        'Instandhaltung': -55000000,
        'Aufwandsentschädigungen': -9000000,
        'Grundsteuer B': -494100,
        'Grundbesitzabgabe': -198600,
        'sonstige Ausgaben': -3000000,
        'Vereinsfest': 0,
        'Satellitenanlage': -5000000,
        'Energieabrechnung': 0,
        'Energieanbieter': 0,
        'Energieabschlag I': 0,
        'Energieabschlag II': 0,
    },
    2019: {
        'Mitgliedsbeitrag': 80250000,
        'Aufnahmegebühr': 250000,
        'Mahn- & Zusatzleistungen': 500000,
        'Arbeitsstunden': 13000000,
        'Körperschaftssteuer': 30000,
        'sonstige Einnahmen': 0,
        'Pacht': 6203900,
        'Verwaltung': -7000000,
        'Versicherungen': -6300000,
        'Müll': -11000000,
        'Abwasser': -600000,
        'Instandhaltung': -55000000,
        'Aufwandsentschädigungen': -8400000,
        'Grundsteuer B': -494100,
        'Grundbesitzabgabe': -198600,
        'sonstige Ausgaben': -3000000,
        'Vereinsfest': 0,
        'Satellitenanlage': -5000000,
        'Energieabrechnung': -2000000,
        'Energieanbieter': 0,
        'Energieabschlag I': 0,
        'Energieabschlag II': 0,
    },
    2020: {
        'Mitgliedsbeitrag': 79500000,
        'Aufnahmegebühr': 500000,
        'Mahn- & Zusatzleistungen': 500000,
        'Arbeitsstunden': 13000000,
        'Körperschaftssteuer': 0,
        'sonstige Einnahmen': 0,
        'Pacht': 6203900,
        'Verwaltung': -7500000,
        'Versicherungen': -6000000,
        'Müll': -12000000,
        'Abwasser': -600000,
        'Instandhaltung': -55000000,
        'Aufwandsentschädigungen': -9000000,
        'Grundsteuer B': -494100,
        'Grundbesitzabgabe': -201100,
        'sonstige Ausgaben': -3000000,
        'Vereinsfest': 0,
        'Satellitenanlage': -5000000,
        'Energieabrechnung': 0,
        'Energieanbieter': 0,
        'Energieabschlag I': 0,
        'Energieabschlag II': 0,
    },
    2021: {
        'Mitgliedsbeitrag': 78750000,
        'Aufnahmegebühr': 500000,
        'Mahn- & Zusatzleistungen': 300000,
        'Arbeitsstunden': 14000000,
        'Körperschaftssteuer': 0,
        'sonstige Einnahmen': 0,
        'Pacht': 6203900,
        'Verwaltung': -7000000,
        'Versicherungen': -8058300,
        'Müll': -12000000,
        'Abwasser': -600000,
        'Instandhaltung': -55000000,
        'Aufwandsentschädigungen': -8400000,
        'Grundsteuer B': -494100,
        'Grundbesitzabgabe': -202100,
        'sonstige Ausgaben': -3000000,
        'Vereinsfest': 0,
        'Satellitenanlage': -5000000,
        'Energieabrechnung': -2000000,
        'Energieanbieter': 0,
        'Energieabschlag I': 0,
        'Energieabschlag II': 0,
    },
}


def upgrade():
    op.create_table(
        'budget',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('organization_id', sa.Integer(), nullable=False),
        sa.Column('value', sa.Integer(), nullable=True),
        sa.Column('accounting_year', sa.Integer(), nullable=True),
        sa.Column('booking_kind_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['booking_kind_id'], ['bookingkind.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    for year, budgets in DATA.items():
        for kind, amount in DATA[year].items():
            id_ = BOOKING_KINDS[kind]
            op.execute(f"""
INSERT into budget (booking_kind_id, accounting_year, value, organization_id)
VALUES ({id_}, {year}, {amount}, 1);
""")


def downgrade():
    op.drop_table('budget')



# encoding=utf8
"""Add new members.

Revision ID: bbb62bf367a
Revises: 2d287284c72e
Create Date: 2015-10-05 09:42:39.792984

"""
from __future__ import unicode_literals
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'bbb62bf367a'
down_revision = '2d287284c72e'


def upgrade():
    op.add_column('electricmeter',
                  sa.Column('discount_to_id', sa.Integer(), nullable=True))

    # Silvia & Lars Müller verkaufen Bungalow 317 an Andrea und Mario Gille
    op.execute("""
        INSERT into member
            (appellation, firstname, lastname, street, zip, city, country,
             email, direct_debit, get_post) VALUES
            ('Frau', 'Andrea', 'Gille', 'Weberweg 8d', '06805',
             'Bad Schmiedeberg', 'Deutschland', 'andrea-gille@t-online.de',
             false, true);""")
    op.execute("""
        UPDATE allotment SET member_id = (
            SELECT id FROM member WHERE lastname = 'Gille') WHERE id = 151;""")

    # Gerd & Irmgard Mank verkaufen Bungalow 250 an Uwe & Silvia Loos
    op.execute("""
        INSERT into member
            (appellation, firstname, lastname, street, zip, city, country,
             direct_debit, get_post) VALUES
            ('Herr', 'Uwe', 'Loos', 'Teucheler Straße 14b', '06886',
             'Wittenberg', 'Deutschland', false, true);""")
    op.execute("""
        UPDATE allotment SET member_id = (
            SELECT id FROM member WHERE lastname = 'Loos') WHERE id = 137;""")

    # Zähler 4246 geht an Bungalow 335 und Abrechnung an Anke Meister
    op.execute("""
        INSERT into member
            (appellation, firstname, lastname, street, zip, city, country,
             direct_debit, get_post, phone, mobile, bic, iban,
             direct_debit_date) VALUES
            ('Frau', 'Anke', 'Meister', 'Bahnhofstr. 13', '06679',
             'Taucha', 'Deutschland', true, true, '034441 92803',
             '0175 1771334', 'GENODEF1NMB', 'DE58800636480220302200',
             '2015-10-05');""")
    op.execute("""UPDATE electricmeter SET
        discount_to_id = (SELECT id FROM member WHERE lastname = 'Meister'),
        allotment_id = 163 WHERE id = 62;""")

    # Petra Tüttmann verkauft Bungalow 255 an Rene Geißler
    op.execute("""
        INSERT into member
            (appellation, firstname, lastname, street, zip, city, country,
             mobile, direct_debit, get_post) VALUES
            ('Herr', 'Rene', 'Geißler', 'Bleddiner Str. 6', '06901',
             'Kember / OT Globig', 'Deutschland', '0152 218976145',
             false, true);""")
    op.execute("""
        UPDATE allotment SET member_id = (
            SELECT id FROM member WHERE lastname = 'Geißler')
        WHERE id = 141;""")


def downgrade():
    op.drop_column('electricmeter', 'discount_to_id')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import datetime
import transaction
import mock


def setUp():
    from sw.allotmentclub import Assignment, AssignmentAttendee, Member
    member = Member.create(firstname='Gerd', lastname='Mittag')
    ass = Assignment.create(day=datetime.datetime(2015, 3, 7, 9),
                            purpose='Frühjahrsputz',
                            responsible=member,
                            accounting_year=2015)
    AssignmentAttendee.create(
        assignment=ass, member=member, hours=4)
    transaction.commit()
    return member, ass


def test_assignments_can_have_attendees(database):
    member, assignment = setUp()
    from sw.allotmentclub import AssignmentAttendee
    assert 1 == len(assignment.attendees)
    assert 1 == len(member.assignments)
    AssignmentAttendee.create(assignment=assignment, member=member, hours=4)
    assert 2 == len(assignment.attendees)
    assert 2 == len(member.assignments)


def test_AssignmentListView_1(browser, json_fixture):
    """It displays list of assignments."""
    url = json_fixture.url()
    setUp()
    browser.login()
    browser.open('http://localhost{}'.format(url))
    json_fixture.assertEqual(browser.json, 'data')


def test_AssignmentAddView_1(browser, json_fixture):
    """It can add new assignments via JSON."""
    from sw.allotmentclub import Assignment
    url = json_fixture.url()
    browser.login()
    with mock.patch('sw.allotmentclub.browser.assignment.datetime_now') as now:
        current_year = datetime.datetime.now().year
        now.return_value = datetime.datetime(current_year, 2, 20, 10)
        browser.post('http://localhost{}'.format(url), data='')
    assert 'success' == browser.json['status']
    json_fixture.assertEqual(browser.json, 'data')
    assert 1 == len(Assignment.query().all())


def test_AssignmentAttendeesEditView_1(browser, json_fixture):
    """It can handle german floats for assignment hours."""
    from sw.allotmentclub import AssignmentAttendee
    url = json_fixture.url()
    setUp()
    browser.login()
    browser.post('http://localhost{}'.format(url), data={'hours': '2,5'})
    assert 'success' == browser.json['status']
    assert 2.5 == AssignmentAttendee.get(1).hours


def test_AssignmentListAttendeesView_1(browser, json_fixture):
    """It can display list of assignment attendees."""
    url = json_fixture.url()
    setUp()
    browser.login()
    browser.open('http://localhost{}'.format(url))
    json_fixture.assertEqual(browser.json, 'data')


def test_AssignmentDeleteAttendeesView_1(browser, json_fixture):
    """It can delete assignment attendees from assignment."""
    url = json_fixture.url('/assignments/1/attendees/1/delete')
    setUp()
    browser.login()
    browser.open('http://localhost{}'.format(url))
    json_fixture.assertEqual(browser.json, 'data')
    browser.open('http://localhost/assignments/1/list')
    assert [] == browser.json_result


def test_AssignmentTodoListView_1(browser, json_fixture):
    """It displays list of assignment todos."""
    url = json_fixture.url()
    setUp()
    browser.login()
    browser.open('http://localhost{}'.format(url))
    json_fixture.assertEqual(browser.json, 'data')


def test_AssignmentTodoAddView_1(browser, json_fixture):
    """It can add new assignment todos via JSON."""
    from sw.allotmentclub import AssignmentTodo
    url = json_fixture.url()
    browser.login()
    browser.post('http://localhost{}'.format(url), data='')
    assert 'success' == browser.json['status']
    json_fixture.assertEqual(browser.json, 'data', save=True)
    assert 1 == len(AssignmentTodo.query().all())
    assert 4 == AssignmentTodo.query().one().priority



"""Add accounting year to assignments.

Revision ID: 50035156da61
Revises: 39a3b6119fc6
Create Date: 2015-04-07 19:34:23.388874

"""

# revision identifiers, used by Alembic.
from alembic import op
import sqlalchemy as sa

revision = '50035156da61'
down_revision = '39a3b6119fc6'


def upgrade():
    op.add_column(
        'assignment',
        sa.Column('accounting_year', sa.Integer(), nullable=True))
    op.execute("""UPDATE assignment SET accounting_year = 2015;""")


def downgrade():
    op.drop_column('assignment', 'accounting_year')



"""Add budget for 2022

Revision ID: 7e0a65cb279f
Revises: e44a6f304d17
Create Date: 2022-01-04 11:30:27.433502

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = '7e0a65cb279f'
down_revision = 'e44a6f304d17'

BOOKING_KINDS = {
    'Mitgliedsbeitrag': 1,
    'Energieabschlag I': 3,
    'Energieabschlag II': 4,
    'Energieabrechnung': 5,
    'Grundsteuer B': 6,
    'Abwasser': 7,
    'Aufnahmegebühr': 13,
    'Grundbesitzabgabe': 14,
    'Versicherungen': 15,
    'Verwaltung': 16,
    'Aufwandsentschädigungen': 17,
    'Instandhaltung': 18,
    'Energieanbieter': 19,
    'Müll': 20,
    'Vereinsfest': 21,
    'Pacht': 22,
    'Körperschaftssteuer': 23,
    'Arbeitsstunden': 2,
    'Mahn- & Zusatzleistungen': 24,
    'sonstige Einnahmen': 25,
    'sonstige Ausgaben': 26,
    'Satellitenanlage': 27,
}

DATA = {
    2022: {
        'Mitgliedsbeitrag': 7800,
        'Aufnahmegebühr': 100,
        'Mahn- & Zusatzleistungen': 30,
        'Arbeitsstunden': 1300,
        'Körperschaftssteuer': 0,
        'sonstige Einnahmen': 0,
        'Pacht': 621,
        'Verwaltung': -1000,
        'Versicherungen': -850,
        'Müll': -1300,
        'Abwasser': -60,
        'Instandhaltung': -2600,
        'Aufwandsentschädigungen': -9900,
        'Grundsteuer B': -50,
        'Grundbesitzabgabe': -21,
        'sonstige Ausgaben': -90,
        'Vereinsfest': -2500,
        'Satellitenanlage': -300,
        'Energieabrechnung': -200,
        'Energieanbieter': 0,
        'Energieabschlag I': 0,
        'Energieabschlag II': 0,
    },
}


def upgrade():
    for year, budgets in DATA.items():
        for kind, amount in DATA[year].items():
            id_ = BOOKING_KINDS[kind]
            amount = amount * 10000
            op.execute(f"""
INSERT into budget (booking_kind_id, accounting_year, value, organization_id)
VALUES ({id_}, {year}, {amount}, 1);
""")


def downgrade():
    for year in DATA:
        op.execute(f"DELETE FROM budget WHERE accounting_year = {year}")



"""Update hours of assignment attendees to support 2.5 hours.

Revision ID: 39a3b6119fc6
Revises: 4d63c477b58f
Create Date: 2015-04-07 12:38:11.475425

"""

# revision identifiers, used by Alembic.
from alembic import op
import sqlalchemy

revision = '39a3b6119fc6'
down_revision = '4d63c477b58f'


def upgrade():
    op.alter_column(
        table_name='assignmentattendee',
        column_name='hours',
        nullable=False,
        type_=sqlalchemy.types.Float(1)
    )


def downgrade():
    op.alter_column(
        table_name='assignmentattendee',
        column_name='hours',
        nullable=False,
        type_=sqlalchemy.types.Integer()
    )



from repoze.sendmail.delivery import MailDataManager
from repoze.sendmail.tests.test_delivery import DummyTransaction
import pytest


@pytest.mark.xfail
def test_abort_w_TPC_does_not_raise_ValueError():
    mdm = MailDataManager(object, (), None)
    txn = DummyTransaction()
    mdm.join_transaction(txn)
    mdm.tpc_phase = 1
    mdm.abort(txn)



"""Add category to depot.

Revision ID: e927a9a287ad
Revises: 7e0a65cb279f
Create Date: 2023-03-22 12:42:43.316748

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'e927a9a287ad'
down_revision = '7e0a65cb279f'


def upgrade():
    op.add_column('depot', sa.Column('category', sa.String(length=100),
                                     nullable=True))


def downgrade():
    op.drop_column('depot', 'category')



# encoding=utf8
from io import BytesIO
from pyramid.response import FileIter
import pybars
from ..log import user_data_log, log_with_user
from .base import date, string_agg, get_selected_year, format_size
from .base import format_mimetype, parse_date
from .letter import render_pdf
from .protocol import format_markdown
from pyramid.view import view_config
from sqlalchemy import func
from sqlalchemy.orm import aliased
from sw.allotmentclub import BankingAccount, BookingKind, MEMBERSHIP_FEE
from sw.allotmentclub import Member, Allotment, Parcel, Booking
from sw.allotmentclub import SaleHistory, MemberAttachment
import collections
import datetime
import sw.allotmentclub.browser.base
import sw.allotmentclub.browser.external


class Query(sw.allotmentclub.browser.base.Query):

    data_class = {
        'Nachname': 'expand'
    }
    data_hide = {
        'Vorname': 'phone,tablet',
        'Flurstück': 'phone,tablet',
        'Mobil': 'phone,tablet',
        'Straße': 'phone,tablet',
        'PLZ': 'phone,tablet',
        'Ort': 'phone,tablet',
        'Telefon': 'phone',
        'E-Mail': 'phone',
        'Bungalow': 'phone',
        'Geburtstag': 'phone',
    }
    formatters = {
        'Geburtstag': date,
    }

    join_condition = (Allotment.member_id == Member.id)

    fields = (
        Member.id.label('#'),
        func.coalesce(
            string_agg(Allotment.number), 'n/a').label('Bungalow'),
        func.coalesce(
            string_agg(Parcel.number), 'n/a').label('Flurstück'),
        Member.lastname.label('Nachname'),
        Member.firstname.label('Vorname'),
        Member.street.label('Straße'),
        Member.zip.label('PLZ'),
        Member.city.label('Ort'),
        Member.phone.label('Telefon'),
        Member.mobile.label('Mobil'),
        Member.email.label('E-Mail'),
        Member.birthday.label('Geburtstag'),
    )

    def select(self):
        return (
            self.db.query(*self.fields)
            .select_from(Member)
            .join(Allotment, self.join_condition)
            .join(Parcel)
            .group_by(Member.id)
            .filter(Member.leaving_year.is_(None))
            .filter(Member.organization_id == self.user.organization_id)
        )


class ActiveQuery(Query):

    def select(self):
        return super(ActiveQuery, self).select().filter(
            Allotment.id.isnot(None))


@view_config(route_name='member_list', renderer='json', permission='view')
class MemberListView(sw.allotmentclub.browser.base.TableView):

    query_class = ActiveQuery
    available_actions = [
        dict(url='member_edit', btn_class='btn-success',
             icon='fa fa-pencil', title='Bearbeiten'),
        dict(url='member_attachment', btn_class='btn-success',
             icon='fa fa-list', title='Anlagen'),
        dict(url='become_member_letter', btn_class='btn-success',
             icon='fa fa-print', title='Beitrittserklärung drucken'),
        dict(url='direct_debit_letter', btn_class='btn-success',
             icon='fa fa-print', title='Lastschrift drucken'),
        dict(url='mv_entrance_list', btn_class='btn-success',
             icon='fa fa-print', title='Einlasslisten drucken'),
        dict(url='member_sale', btn_class='btn-success',
             icon='fa fa-money', title='Verkauf'),
        dict(url='member_sale_history', btn_class='btn-success',
             icon='fa fa-money', title='Verkaufhistorie'),
        dict(url='membership_fee', btn_class='btn-danger',
             icon='fa fa-money', title='Mitgliedsbeiträge generieren')]


class PassiveQuery(Query):

    join_condition = (Member.passive_allotment_id == Allotment.id)

    def select(self):
        passive = super(PassiveQuery, self).select().filter(
            Member.passive_allotment_id.isnot(None)
        )
        other = (
            self.db.query(*self.fields)
            .select_from(Member)
            .outerjoin(Allotment, Allotment.member_id == Member.id)
            .outerjoin(Parcel)
            .group_by(Member.id)
            .filter(Member.leaving_year.is_(None))
            .filter(Allotment.member_id.is_(None))
            .filter(Member.passive_allotment_id.is_(None))
            .filter(Member.organization_id == self.user.organization_id)
        )
        return passive.union(other).distinct()


@view_config(route_name='member_list_passive', renderer='json',
             permission='view')
class MemberListPassiveView(MemberListView):

    query_class = PassiveQuery
    available_actions = [
        dict(url='member_add', btn_class='btn-success',
             icon='fa fa-plus', title='Neu'),
        dict(url='member_edit', btn_class='btn-success',
             icon='fa fa-pencil', title='Bearbeiten'),
        dict(url='member_attachment', btn_class='btn-success',
             icon='fa fa-list', title='Anlagen'),
        dict(url='become_member_letter', btn_class='btn-success',
             icon='fa fa-print', title='Beitrittserklärung drucken'),
        dict(url='direct_debit_letter', btn_class='btn-success',
             icon='fa fa-print', title='Lastschrift drucken'),
    ]


class LeasedQuery(Query):

    def select(self):
        return super(LeasedQuery, self).select().filter(
            Parcel.leased.is_(True))


@view_config(route_name='member_list_leased', renderer='json',
             permission='view')
class MemberListLeasedView(MemberListView):

    query_class = LeasedQuery
    available_actions = []


class TapWaterQuery(Query):

    def select(self):
        return super(TapWaterQuery, self).select().filter(
            Parcel.tap_water.is_(True))


@view_config(route_name='member_list_tap_water', renderer='json',
             permission='view')
class MemberListTapWaterView(MemberListView):

    query_class = TapWaterQuery
    available_actions = []


@view_config(route_name='member_edit', renderer='json', permission='view')
class MemberEditView(
        sw.allotmentclub.browser.external.ExternalRecipientEditView):

    title = 'Mitglied bearbeiten'

    @property
    def load_options(self):
        options = super(MemberEditView, self).load_options
        options.update({
            'mobile': {'label': 'Mobil-Telefon'},
            'birthday': {
                'label': 'Geburtstag',
                'css_class': 'datepicker',
            },
            'direct_debit': {'label': 'Lastschrift',
                             'template': 'form_boolean'},
            'direct_debit_account_holder': {
                'label': 'Abweichender Kontoinhaber'},
            'iban': {'label': 'IBAN'},
            'bic': {'label': 'BIC'},
            'direct_debit_date': {
                'label': 'Lastschrift-Datum',
                'css_class': 'datepicker'},
            'get_post': {'label': 'Postversand?', 'template': 'form_boolean'},
            'note': {'label': 'Hinweis', 'template': 'form_markdown'},
        })
        if not self.context.active:
            options.update({
                'passive_allotment_id': {
                    'label': 'Passive Mitgliedschaft',
                    'source': self.allotment_source,
                    'css_class': 'chosen',
                    'required': False
                }
            })
        return options

    @property
    def load_data(self):
        fields = list(super(MemberEditView, self).load_data.items())
        fields.extend([
            ('phone', self.context.phone),
            ('mobile', self.context.mobile),
            ('birthday', date(self.context.birthday)),
            ('direct_debit', self.context.direct_debit),
            ('direct_debit_account_holder',
                self.context.direct_debit_account_holder),
            ('iban', self.context.iban),
            ('bic', self.context.bic),
            ('direct_debit_date', date(self.context.direct_debit_date)),
            ('get_post', self.context.get_post),
        ])
        if not self.context.active:
            fields.append(
                ('passive_allotment_id', self.context.passive_allotment_id)
            )
        fields.append(('note', self.context.note))
        return collections.OrderedDict(fields)

    def save(self, key, value):
        if key in ('direct_debit_date', 'birthday'):
            if value:
                value = parse_date(value).date()
            else:
                value = None
        return super(MemberEditView, self).save(key, value)


@view_config(route_name='member_add', renderer='json', permission='view')
class MemberAddView(MemberEditView):

    def __init__(self, context, request):
        context = Member.create()
        context.commit()
        super(MemberAddView, self).__init__(context, request)
        log_with_user(
            user_data_log.info, self.request.user,
            'Mitglied %s hinzugefügt.', self.context.id)

    @property
    def route_name(self):
        return 'member_edit'


def get_membership_fees(year=None, currency_formatter=None):
    if year is None:
        year = get_selected_year()
    members = {}
    for parcel in Parcel.query().all():
        member = parcel.allotment.member
        if not member:
            continue
        if member.lastname in ['Verein', 'Werkstatt']:
            continue
        if member not in members:
            members[member] = [parcel]
        elif parcel not in members[member]:
            members[member].append(parcel)
    result = {}
    for member, parcels in members.items():
        parcel_numbers = '/'.join(
            sorted(set([str(p.number) for p in parcels])))
        parcel_size = sum(p.size for p in parcels if p.size)
        for k, v in sorted(MEMBERSHIP_FEE.items()):
            if parcel_size <= k:
                fee = v
                break
        if currency_formatter is not None:
            fee = currency_formatter(fee)
        result[member] = dict(
            parcel_number=parcel_numbers,
            parcel_size=parcel_size,
            membership_fee=fee
        )
    return result


@view_config(route_name='membership_fee', renderer='json', permission='view')
class MembershipFeeView(sw.allotmentclub.browser.base.View):

    def update(self):
        account = (BankingAccount.query()
                   .filter(BankingAccount.number == '1')).one()
        kind = (BookingKind.query()
                .filter(BookingKind.title == 'Mitgliedsbeitrag')).one()
        year = get_selected_year()
        for member, fee_data in get_membership_fees(year).items():
            parcel_numbers = fee_data['parcel_number']
            parcel_size = fee_data['parcel_size']
            fee = fee_data['membership_fee']
            Booking.create(
                banking_account=account,
                purpose=(
                    f'Mitgliedsbeitrag {year} Flurstück {parcel_numbers} '
                    f'({parcel_size}qm)'
                ),
                value=0 - fee,
                accounting_year=year,
                booking_day=datetime.date(int(year), 3, 31),
                member=member,
                kind=kind)


class MemberLetter(sw.allotmentclub.browser.base.AddView):

    def get_pdf(self):
        subject = self.request.params['subject']
        message = format_markdown(self.request.params['message'])
        attachments = self.request.params.get('attachments')
        member = self.context if self.with_address else None
        return render_pdf(member, subject, message, self.request.user,
                          attachments=attachments)


def get_member_form_data(member):
    allotments = member.allotments
    if not allotments:
        allotments = member.passive_allotment
    allotment = '/'.join(str(a.number) for a in allotments)

    data = {'member': [
        {'key': 'Anrede', 'value': member.appellation},
        {'key': 'Titel', 'value': member.title},
        {'key': 'Vorname', 'value': member.firstname},
        {'key': 'Nachname', 'value': member.lastname},
        {'key': 'Bungalownummer', 'value': allotment},
        {'key': 'Strasse', 'value': member.street},
        {'key': 'PLZ', 'value': member.zip},
        {'key': 'Ort', 'value': member.city},
        {'key': 'Telefon', 'value': member.phone},
        {'key': 'Mobiltelefon', 'value': member.mobile},
        {'key': 'E-Mail-Adresse', 'value': member.email},
        {'key': 'Geburtstag', 'value': date(member.birthday)},
        {'key': 'IBAN', 'value': member.iban},
        {'key': 'BIC', 'value': member.bic},
        {'key': 'Name Kreditinstitut', 'value': ''},
    ]}

    for item in data['member']:
        if not item['value']:
            item['value'] = '&nbsp;'
    return data


@view_config(route_name='direct_debit_letter', renderer='json',
             permission='view')
class DirectDebitLetter(sw.allotmentclub.browser.base.PrintBaseView):

    with_address = None
    filename = 'Lastschrift'
    subject = 'Lastschrifteinzugsermächtigung'
    subsubject = 'Gläubiger-ID: <b>DE42ZZZ00000348413</b>'
    intro = ''
    message = """
<table style="font-size: 10pt;">
  <tbody>
    {{#each member}}
    <tr>
      <td style="width: 30%;">{{key}}: </td>
      <td style="width: 70%; height: 25pt;">
        {{{value}}}
        <hr style="margin-top: 0px;" />
      </td>
    </tr>
    {{/each}}
  </tbody>
</table>
<p>
Der Vorstand der Bungalowgemeinschaft Roter See e.V. wird hiermit widerruflich
ab sofort ermächtigt, alle im Zusammenhang mit meiner Mitgliedschaft
bestehenden Forderungen wie z.B. Mitgliedsbeiträge oder Energiekosten mittels
Lastschriftverfahren von meinem Konto einzuziehen.
</p>
<table style="font-size: 10pt;">
  <tbody>
    <tr>
      <td style="width: 20%">&nbsp;<hr/>Datum</td>
      <td style="width: 40%;">&nbsp;<hr/>Ort</td>
      <td style="width: 40%;">&nbsp;<hr/>Unterschrift Kontoinhaber</td>
    </tr>
  </tbody>
</table>"""

    def get_pdf(self):
        subject = self.subject
        compiler = pybars.Compiler()
        template = compiler.compile(self.message)
        data = get_member_form_data(self.context)

        message = "".join(template(data))
        member = self.context if self.with_address else None
        return render_pdf(member, subject, message, self.request.user,
                          subsubject=self.subsubject)


@view_config(route_name='become_member_letter', renderer='json',
             permission='view')
class BecomeMemberLetter(sw.allotmentclub.browser.base.PrintBaseView):

    with_address = None
    subject = 'Beitrittserklärung'
    subsubject = ''
    intro = ''
    filename = 'Beitritt'
    message = """
<table style="font-size: 10pt;">
  <tbody>
    {{#each member}}
    <tr>
      <td style="width: 30%;">{{key}}: </td>
      <td style="width: 70%; height: 25pt;">
        {{{value}}}
        <hr style="margin-top: 0px;" />
      </td>
    </tr>
    {{/each}}
  </tbody>
</table>
<p>
Hiermit erkläre ich meinen Beitritt zum Verein Leuna-Bungalowgemeinschaft
"Roter See" e.V. und erkenne die mir vorgelegte Satzung und die Ordnungen der
Leuna-Bungalowgemeinschaft an.
</p>
<table style="font-size: 10pt;">
  <tbody>
    <tr>
      <td style="width: 20%">&nbsp;<hr/>Datum</td>
      <td style="width: 40%;">&nbsp;<hr/>Ort</td>
      <td style="width: 40%;">&nbsp;<hr/>Unterschrift</td>
    </tr>
  </tbody>
</table>"""

    def get_pdf(self):
        subject = self.subject
        compiler = pybars.Compiler()
        template = compiler.compile(self.message)
        data = get_member_form_data(self.context)
        data['member'] = data['member'][:-3]

        message = "".join(template(data))
        member = self.context if self.with_address else None
        return render_pdf(member, subject, message, self.request.user,
                          subsubject=self.subsubject)


@view_config(route_name='member_sale', renderer='json', permission='view')
class MemberSaleView(sw.allotmentclub.browser.base.EditJSFormView):

    title = 'Grundstücks-Verkauf'

    @property
    def load_options(self):
        return {
            'buyer_id': {
                'source': self.member_source,
                'label': 'Käufer',
                'css_class': 'chosen'
            },
            'transfer_member_due': {
                'label': 'Mitgliedsbeitrag übernehmen',
                'template': 'form_boolean'
            },
            'transfer_advance_pay_one': {
                'label': 'Abschlag I übernehmen',
                'template': 'form_boolean'
            },
            'transfer_advance_pay_two': {
                'label': 'Abschlag II übernehmen',
                'template': 'form_boolean'
            },
            'transfer_energy_bill': {
                'label': 'Energieabrechnung übernehmen',
                'template': 'form_boolean'
            },
            'transfer_assignment_due': {
                'label': 'Arbeitsstunden übernehmen',
                'template': 'form_boolean'
            }
        }

    @property
    def load_data(self):
        fields = [
            ('buyer_id', self.context.id),
            ('transfer_member_due', False),
            ('transfer_advance_pay_one', False),
            ('transfer_advance_pay_two', False),
            ('transfer_energy_bill', False),
            ('transfer_assignment_due', False)]
        return collections.OrderedDict(fields)

    def save(self, key, value):
        year = get_selected_year()
        if key == 'buyer_id':
            buyer = Member.get(value)
            for allotment in self.context.allotments:
                allotment.member = buyer
            SaleHistory.create(seller_id=self.context.id, buyer_id=buyer.id)
            self.context.leaving_year = year
            Booking.create(
                banking_account_id=2,
                booking_day=datetime.date.today(),
                purpose='Aufnahmebeitrag',
                value=-250000,
                member_id=buyer.id,
                kind_id=13,
                accounting_year=year)
        else:
            sale = (SaleHistory.query()
                    .filter(SaleHistory.date == datetime.date.today())
                    .filter(SaleHistory.seller_id == self.context.id)).one()
            buyer = Member.get(sale.buyer_id)
            query = dict(
                transfer_member_due='Mitgliedsbeitrag {}'.format(year),
                transfer_advance_pay_one='Energieabschlag I',
                transfer_advance_pay_two='Energieabschlag II',
                transfer_energy_bill='Energieabrechnung',
                transfer_assignment_due='Arbeitsstunden {}'.format(year)
            )[key]
            for booking in (
                    Booking.query()
                    .filter(Booking.accounting_year == year)
                    .filter(Booking.purpose.ilike('%{}%'.format(query)))
                    .filter(Booking.member == self.context)).all():
                booking.member = buyer
        return True


@view_config(route_name='mv_entrance_list', renderer='json', permission='view')
class MVEntranceListView(sw.allotmentclub.browser.base.XLSXExporterView):

    filename = 'Einlassliste'
    title = 'Liste Einlass MV'

    def query(self):
        return (
            self.db.query(
                Member.lastname.label('Nachname'),
                Member.firstname.label('Vorname'),
                string_agg(Allotment.number).label('Bungalow'),
                Member.email.label('E-Mail-Adresse'),
                Member.direct_debit.label('Lastschrift'),
                Member.id.label('Unterschrift'),
                Member.id.label('Vollmacht'),
            )
            .select_from(Member)
            .join(Allotment, Member.id == Allotment.member_id)
            .group_by(Member.id)
            .filter(Member.lastname != 'Werkstatt')
            .filter(Member.lastname != 'Verein')
            .order_by(Member.lastname)
        )

    def _data(self, query):
        data = super(MVEntranceListView, self)._data(query)
        for item in data:
            item = list(item)
            item[-1] = item[-2] = ''
            yield item


seller = aliased(Member, name='seller')
buyer = aliased(Member, name='buyer')


class MemberSaleHistoryQuery(sw.allotmentclub.browser.base.Query):

    data_class = {
        'Bungalow': 'expand'
    }
    data_hide = {
        'Käufer': 'phone',
        'Datum': 'phone',
        'Verkäufer': 'phone',
        'Flurstück': 'phone,table',
    }

    def select(self):
        return (
            self.db.query(
                SaleHistory.id.label('#'),
                SaleHistory.date.label('Datum'),
                func.coalesce(
                    string_agg(Allotment.number), 'n/a').label('Bungalow'),
                func.coalesce(
                    string_agg(Parcel.number), 'n/a').label('Flurstück'),
                (seller.lastname + ', ' + seller.firstname).label('Verkäufer'),
                (buyer.lastname + ', ' + buyer.firstname).label('Käufer'),
            )
            .select_from(SaleHistory)
            .join(seller, SaleHistory.seller_id == seller.id)
            .join(buyer, SaleHistory.buyer_id == buyer.id)
            .outerjoin(
                Allotment,
                Allotment.member_id.in_((seller.id, buyer.id))
            )
            .outerjoin(Parcel)
            .group_by(SaleHistory.id, seller.lastname, seller.firstname,
                      buyer.lastname, buyer.firstname)
        )


@view_config(route_name='member_sale_history', renderer='json',
             permission='view')
class MemberSaleHistoryView(sw.allotmentclub.browser.base.TableView):

    query_class = MemberSaleHistoryQuery
    default_order_by = 'date'


class AttachmentQuery(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Größe': format_size,
        'Dateityp': format_mimetype,
    }
    data_class = {
        'Name': 'expand'
    }
    data_hide = {
        'Dateityp': 'phone,tablet',
        'Größe': 'phone,tablet',
    }

    def select(self):
        return (
            self.db.query(
                MemberAttachment.id.label('#'),
                MemberAttachment.name.label('Name'),
                MemberAttachment.mimetype.label('Dateityp'),
                MemberAttachment.size.label('Größe'))
            .select_from(MemberAttachment)
            .filter_by(member_id=self.context.id))


@view_config(route_name='member_attachment', renderer='json',
             permission='view')
class MemberAttachmentsView(sw.allotmentclub.browser.base.TableView):
    """ Liste aller Anlagen."""

    query_class = AttachmentQuery
    default_order_by = 'Name'
    available_actions = [
        dict(url='member_attachment_add', btn_class='btn-success',
             icon='fa fa-plus', title='Neu'),
        dict(url='member_attachment_delete', btn_class='btn-danger',
             icon='glyphicon glyphicon-trash', title='Löschen'),
        dict(url='member_attachment_download', btn_class='btn-success',
             icon='fa fa-download', title='Herunterladen')]


@view_config(route_name='member_attachment_add', renderer='json',
             permission='view')
class MemberAttachmentAddView(sw.allotmentclub.browser.base.AddView):

    model = MemberAttachment

    def parse_file(self):
        file = self.request.params.get('file')
        name = file.filename
        data = file.file
        data.seek(0)
        data = data.read()
        mimetype = file.type
        size = len(data)
        return name, mimetype, size, data

    def __call__(self):
        if not self.form_submit():
            return {'status': 'success', 'data': {}}
        name, mimetype, size, data = self.parse_file()
        attachment = self.model.create(
            name=name, mimetype=mimetype, size=size, data=data,
            member_id=self.context.id)
        attachment.commit()
        log_with_user(user_data_log.info,
                      self.request.user,
                      'hat Anlage %s zu Mitglied %s hinzugefügt.',
                      attachment.id, self.context.id)
        return {'status': 'success'}


@view_config(route_name='member_attachment_download',
             permission='view')
class MemberAttachmentDownloadView(sw.allotmentclub.browser.base.View):

    def __call__(self):
        response = self.request.response
        response.set_cookie('fileDownload', value='true')
        response.content_type = self.context.mimetype
        response.content_length = int(self.context.size)
        response.content_disposition = (
            'attachment; filename="{}"'.format(self.context.name))
        response.app_iter = FileIter(BytesIO(self.context.data))
        return response


@view_config(route_name='member_attachment_delete', renderer='json',
             permission='view')
class MemberAttachmentDeleteView(sw.allotmentclub.browser.base.DeleteView):

    model = MemberAttachment

    def log(self):
        if self.deleted is not None:
            deleted = self.context.id
            member = self.context.member_id
            log_with_user(user_data_log.info,
                          self.request.user,
                          'hat Anlage %s von Mitglied %s gelöscht.',
                          deleted, member)



# encoding=utf-8
from ..log import app_log
from .base import string_agg, date, boolean, route_url
from .depot import DepotBase
from datetime import datetime
from io import BytesIO
from pyramid.response import FileIter
from pyramid.response import Response
from pyramid.view import view_config
from reportlab.graphics import renderPDF
from reportlab.pdfgen.canvas import Canvas
from svglib.svglib import svg2rlg
from sw.allotmentclub import Parcel, Allotment, Member
import os
import pkg_resources
import sw.allotmentclub.browser.base
import tempfile


class Query(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Karte hinterlegt': boolean,
    }

    def select(self):
        return (
            self.db.query(
                Parcel.id.label('#'),
                Parcel.number.label('Flurstück'),
                string_agg(Allotment.number).label('Bungalow'),
                Parcel.size.label('Größe in qm'),
                Member.lastname.label('Nachname'),
                Member.firstname.label('Vorname'),
                Parcel.map_mimetype.label('Karte hinterlegt'),
            )
            .select_from(Parcel)
            .join(Allotment)
            .join(Member, Member.id == Allotment.member_id)
            .group_by(
                Parcel.id,
                Parcel.number,
                Parcel.size,
                Member.firstname,
                Member.lastname,
                Parcel.map_mimetype,
            ))


@view_config(route_name='map', renderer='json', permission='view')
class MapView(sw.allotmentclub.browser.base.TableView):

    query_class = Query
    available_actions = [
        dict(url='map_download', btn_class='btn-success',
             icon='fa fa-download', title='Herunterladen')]

    def __call__(self):
        result = super(MapView, self).__call__()
        map_data = {}
        for _, parcel, allotment, _, ln, fn, _ in sw.allotmentclub.json_result(
                result['data']['data']):
            if fn:
                ln += ', {}.'.format(fn[0])
            if str(parcel) not in map_data:
                map_data[str(parcel)] = [str(allotment), ln]
            elif map_data[str(parcel)][1] == ln:
                map_data[str(parcel)] = [
                    map_data[str(parcel)][0] + f', {allotment}',
                    map_data[str(parcel)][1]
                ]
            else:
                map_data[str(parcel)] = [
                    map_data[str(parcel)][0] + f', {allotment}',
                    map_data[str(parcel)][1] + f', {ln}'
                ]
        result['data']['map_data'] = map_data
        with open(pkg_resources.resource_filename(
                'sw.allotmentclub.browser', 'lageplan.svg')) as map_:
            result['data']['map'] = map_.read()
        return result


@view_config(route_name='map_download', permission='view')
class MapDownloadView(object):

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self):
        data = self.request.POST['svg']
        msg = 'Lageplan vom %s' % date(datetime.now())

        data = data.encode('utf-8')
        return Response(
            content_type='image/svg+xml',
            content_disposition='attachment; filename="Lageplan.svg"',
            content_length=len(data),
            charset='utf-8',
            body=data
        )

        svg_handle, svg_filename = tempfile.mkstemp(suffix='svg')
        try:
            os.fdopen(svg_handle, 'wb').write(data)
            app_log.info('Map written to `{}`.'.format(svg_filename))
            d = renderPDF.renderScaledDrawing(svg2rlg(svg_filename))
            output = BytesIO()
            c = Canvas(output)
            c.setTitle(msg)
            c.setPageSize((d.width, d.height))
            renderPDF.draw(d, c, 0, 0)
            c.showPage()
            c.save()
        finally:
            os.remove(svg_filename)
        output.seek(0)
        data = output.read()
        return Response(
            content_type='application/pdf',
            content_disposition='attachment; filename="Lageplan.pdf"',
            content_length=len(data),
            charset=b'utf-8',
            body=data)


@view_config(route_name='parcel_list', renderer='json', permission='view')
class ParcelListView(sw.allotmentclub.browser.base.TableView):

    query_class = Query
    available_actions = [
        dict(url='parcel_map_upload', btn_class='btn-success',
             icon='fa fa-upload', title='Hinterlegen'),
        dict(url='parcel_map_download_check', btn_class='btn-success',
             icon='fa fa-download', title='Herunterladen')]


@view_config(route_name='parcel_map_upload', renderer='json',
             permission='view')
class ParcelMapAddView(DepotBase):

    action = 'eine Karte hinzugefügt'

    def __call__(self):
        if not self.form_submit():
            return {'status': 'success', 'data': {}}
        name, mimetype, size, data, date, user = self.parse_file()
        self.context.map_mimetype = mimetype
        self.context.map_size = size
        self.context.map_data = data
        self.log(self.context.id)
        return {'status': 'success'}


@view_config(
    route_name='parcel_map_download_check',
    permission='view',
    renderer='json')
class ParcelMapDownloadCheckView(sw.allotmentclub.browser.base.View):

    def update(self):
        if not self.context.map_mimetype:
            self.result = {
                'status': 'error',
                'message': 'Kein Anhang zum Download vorhanden'
            }
        else:
            download_url = (
                '/api' + route_url('parcel_map_download', self.request))
            download_url = download_url.replace('{id}', str(self.context.id))
            self.result = {
                'status': 'success',
                'redirect': download_url
            }


@view_config(route_name='parcel_map_download', permission='view')
class ParcelMapDownloadView(object):

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self):
        response = self.request.response
        response.set_cookie('fileDownload', value='true')
        response.content_type = self.context.map_mimetype
        response.content_length = int(self.context.map_size)
        response.content_disposition = 'attachment; filename=%s_map.%s' % (
            self.context.number,
            self.context.map_mimetype.split('/')[1]
        )
        response.app_iter = FileIter(BytesIO(self.context.map_data))
        return response



"""Model for splitting bookings.

Revision ID: 43123080ac3a
Revises: 568f7817789f
Create Date: 2015-03-11 10:12:44.584094

"""

# revision identifiers, used by Alembic.
from alembic import op
import sqlalchemy as sa

revision = '43123080ac3a'
down_revision = '568f7817789f'


def upgrade():
    op.add_column(
        'booking',
        sa.Column('is_splitted', sa.Boolean(), nullable=False,
                  server_default='False'))
    op.add_column(
        'booking', sa.Column('splitted_from_id', sa.Integer(), nullable=True))


def downgrade():
    op.drop_column('booking', 'splitted_from_id')
    op.drop_column('booking', 'is_splitted')



# encoding=utf-8
import hashlib
from sw.allotmentclub import Member, Organization, User
from sw.allotmentclub.account import import_transactions_from_fints
from sw.allotmentclub.application import Application
import risclog.sqlalchemy.interfaces
import csv
import datetime
import gocept.logging
import os
import pyramid.testing
import pyramid.threadlocal
import transaction
import zope.component


ORGANIZATION_ID = 1


def import_members():
    parser = gocept.logging.ArgumentParser(
        description="Import members from excel file.")
    parser.add_argument('-c', '--config', default='portal.ini',
                        help='Specify the config file. (default: portal.ini)')
    parser.add_argument(
        '-i', '--input', default='members.xslx',
        help='Specify the input file. (default: members.xslx)')
    options = parser.parse_args()
    app = Application.from_filename(options.config)
    registry = pyramid.registry.Registry(
        bases=(zope.component.getGlobalSiteManager(),))
    config = pyramid.config.Configurator(
        settings=app.settings, registry=registry)
    config.setup_registry(settings=app.settings)
    request = pyramid.testing.DummyRequest(_registry=registry)
    request.client_addr = '127.0.0.1'
    context = pyramid.threadlocal.manager.get().copy()
    context['request'] = request
    context['registry'] = registry
    pyramid.threadlocal.manager.push(context)
    with open(options.input, 'r') as file:
        reader = csv.reader(file, delimiter=';')
        for id, line in enumerate(reader):
            if id == 0:
                continue
            Member.find_or_create(
                title='',
                appellation='Herr',
                organization_id=2,
                lastname=line[0],
                firstname=line[1],
                street=line[2],
                zip=line[3],
                city=line[4],
                country='Deutschland',
                phone=line[5],
                mobile=line[6],
                email=line[7],
                birthday=line[8])
    transaction.commit()


def import_transactions():
    parser = gocept.logging.ArgumentParser(
        description="Import banking transactions.")
    parser.add_argument('-c', '--config', default='portal.ini',
                        help='Specify the config file. (default: portal.ini)')
    options = parser.parse_args()
    app = Application.from_filename(options.config)
    registry = pyramid.registry.Registry(
        bases=(zope.component.getGlobalSiteManager(),))
    config = pyramid.config.Configurator(
        settings=app.settings, registry=registry)
    config.setup_registry(settings=app.settings)
    request = pyramid.testing.DummyRequest(_registry=registry)
    request.client_addr = '127.0.0.1'
    context = pyramid.threadlocal.manager.get().copy()
    context['request'] = request
    context['registry'] = registry
    pyramid.threadlocal.manager.push(context)
    user = User.by_username('system')
    import_transactions_from_fints(user)


VCF_FILES = {
    1: 'sweh',
    2: 'blauerball'
}


def export_members_vcf():
    parser = gocept.logging.ArgumentParser(
        description="Export members in vcf format.")
    parser.add_argument('-c', '--config', default='portal.ini',
                        help='Specify the config file. (default: portal.ini)')
    options = parser.parse_args()
    Application.from_filename(options.config)
    for organization in Organization.query().all():
        output = ""
        for m in (
                Member.query()
                .filter(Member.organization_id == organization.id)
                .filter(Member.leaving_year.is_(None))):
            output += """BEGIN:VCARD
VERSION:3.0
PRODID:-//Sebastian Wehrmann//sw.allotmentclub//DE
N:{lastname};{firstname};;;
FN:{firstname} {lastname}
ADR;type=HOME;type=pref:;;{street};{city};;{zip};{country}""".format(
                lastname=m.lastname, firstname=m.firstname, street=m.street,
                city=m.city, zip=m.zip, country=m.country)
            if m.email:
                output += """
EMAIL;type=INTERNET;type=HOME;type=pref:{}""".format(m.email)
            if m.mobile:
                output += """
TEL;type=CELL;type=VOICE;type=pref:{}""".format(m.mobile)
            if m.phone:
                output += """
TEL;type=HOME;type=VOICE:{}""".format(m.phone)
            if m.birthday:
                output += """
BDAY:{}""".format(m.birthday.strftime('%Y%m%d'))
            if m.allotments:
                output += """
NOTE:Bungalow: {}""".format('/'.join(str(a.number) for a in m.allotments))
            elif m.passive_allotment:
                output += """
NOTE:Bungalow: {}""".format(
                    '/'.join(str(a.number) for a in m.passive_allotment))
            output += """
ORG:{org}
REV:{rev}""".format(org=organization.title,
                    rev=datetime.datetime.now().isoformat())
            output += """
UID:{uid}
X-RADICALE-NAME:{uid}.vcf
END:VCARD
""".format(uid=m.id)
        path = os.path.join(os.path.expanduser("~"),
                            '.config/radicale/collections',
                            VCF_FILES[organization.id])
        if not os.path.exists(path):
            os.makedirs(path)
        with open(os.path.join(path, 'addressbook.vcf'), 'w') as f:
            f.write(output)


def export_events_ics():
    import sw.allotmentclub.browser.calendar
    parser = gocept.logging.ArgumentParser(
        description="Export events in ics format.")
    parser.add_argument('-c', '--config', default='portal.ini',
                        help='Specify the config file. (default: portal.ini)')
    parser.add_argument('-t', '--type', help='The type (Mitglied|Vorstand)')
    options = parser.parse_args()
    assert options.type in ('Mitglied', 'Vorstand')

    queries = {
        'Mitglied': sw.allotmentclub.browser.calendar.MitgliederQuery,
        'Vorstand': sw.allotmentclub.browser.calendar.VorstandQuery
    }

    Application.from_filename(options.config)

    name = (
        'Mitgliederliste' if options.type == 'Mitglied' else 'Vorstandsliste'
    )
    output = f"""\
BEGIN:VCALENDAR
VERSION:2.0
PRODID:VerwaltungLeunaBundalow
X-WR-CALNAME:Leuna-Siedlung {name}
NAME:Leuna-Siedlung Mitgliederliste
CALSCALE:GREGORIAN
BEGIN:VTIMEZONE
TZID:Europe/Berlin
TZURL:http://tzurl.org/zoneinfo-outlook/Europe/Berlin
X-LIC-LOCATION:Europe/Berlin
BEGIN:DAYLIGHT
TZOFFSETFROM:+0100
TZOFFSETTO:+0200
TZNAME:CEST
DTSTART:19700329T020000
RRULE:FREQ=YEARLY;BYMONTH=3;BYDAY=-1SU
END:DAYLIGHT
BEGIN:STANDARD
TZOFFSETFROM:+0200
TZOFFSETTO:+0100
TZNAME:CET
DTSTART:19701025T030000
RRULE:FREQ=YEARLY;BYMONTH=10;BYDAY=-1SU
END:STANDARD
END:VTIMEZONE
"""

    db = zope.component.getUtility(
        risclog.sqlalchemy.interfaces.IDatabase
    )
    query = queries[options.type](db, User.by_username('system')).select()
    for event in query:
        m = hashlib.md5()
        m.update('-'.join(str(i) for i in event).encode())
        uid = m.hexdigest()
        if event[-2]:
            timestamp = event[4].strftime('%Y%m%dT000000Z')
            start = end = event[4].strftime('%Y%m%d')
            if event[5]:
                end = event[5].strftime('%Y%m%d')
        else:
            timestamp = event[4].strftime('%Y%m%dT%H%M%SZ')
            start = end = event[4].strftime('%Y%m%dT%H%M%S')
            if event[5]:
                end = event[5].strftime('%Y%m%dT%H%M%S')
        output += f"""\
BEGIN:VEVENT
DTSTAMP:{timestamp}
UID:{uid}@roter-see.de
DTSTART;TZID=Europe/Berlin:{start}
DTEND;TZID=Europe/Berlin:{end}
SUMMARY:{event[2]}
LOCATION:{event[3]}
END:VEVENT
"""
    output += "END:VCALENDAR"
    print(output)



"""Replace electric meter for Verein (Zählerwechsel)

Revision ID: ad2b49f5d70a
Revises: d0bfe38ddce2
Create Date: 2021-10-03 09:40:38.096547

"""
from alembic import op
import risclog.sqlalchemy

# revision identifiers, used by Alembic.
revision = 'ad2b49f5d70a'
down_revision = 'd0bfe38ddce2'

OLD_NUMBER = '409120'
OLD_VALUE = 5386
NEW_NUMBER = '6002'
NEW_VALUE = 0
YEAR = 2021


def upgrade():
    from sw.allotmentclub import ElectricMeter
    import transaction

    db = risclog.sqlalchemy.db.get_database()
    db._engines['portal'] = dict(
        engine=op.get_bind().engine,
        alembic_location="sw.allotmentclub:portal",
    )

    old = (ElectricMeter.query()
           .filter(ElectricMeter.number == OLD_NUMBER)
           .one())
    old.replace(old_value=OLD_VALUE, new_number=NEW_NUMBER,
                new_value=NEW_VALUE)
    transaction.commit()


def downgrade():
    from sw.allotmentclub import ElectricMeter
    import transaction

    db = risclog.sqlalchemy.db.get_database()
    db._engines['portal'] = dict(
        engine=op.get_bind().engine,
        alembic_location="sw.allotmentclub:portal",
    )

    to_delete = (ElectricMeter.query()
                 .filter(ElectricMeter.number == NEW_NUMBER)
                 .one())
    to_delete.get_value(YEAR).delete()
    to_delete.delete()

    old = (ElectricMeter.query()
           .filter(ElectricMeter.number == OLD_NUMBER)
           .one())
    old.get_value(YEAR-1).delete()
    transaction.commit()



"""Add birthday to member.

Revision ID: 78f2790c1501
Revises: 3b50f99a2e2f
Create Date: 2017-02-07 09:42:06.154891

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '78f2790c1501'
down_revision = '3b50f99a2e2f'


def upgrade():
    op.add_column('member', sa.Column('birthday', sa.Date(), nullable=True))


def downgrade():
    op.drop_column('member', 'birthday')



from .model import Object
import decimal
import re
import warnings

# VALUE_PER_MEMBER = 600000  # until end of 2015
# VALUE_PER_MEMBER = 750000  # until end of 2016
# VALUE_PER_MEMBER = 650000  # until end of 2018
# VALUE_PER_MEMBER = 750000  # until end of 2021

MEMBERSHIP_FEE = {
    300: 700000,
    500: 750000,
    750: 850000,
    1000: 950000,
    99999: 1050000,
    }

SCAN_IGNORE_TESTS = [re.compile(r'\.testing$').search,
                     re.compile(r'\.conftest$').search,
                     re.compile(r'\.tests\.').search]

warnings.simplefilter('ignore')  # suppress sqlalchemy warnings


# Round for German expetations, e. g. 12.305 to 12.31:
decimal.DefaultContext.rounding = decimal.ROUND_HALF_UP


def json_result(data_list):
    """Remove meta information like `css_class` from result lists."""
    result = []
    for data in data_list:
        result.append([d['value'] for d in data])
    return result


# Initialize SQLAlchemy models
for module in ['model', 'user', 'bulletins', 'assignment', 'depot', 'protocol',
               'electricity', 'account', 'mail', 'keylist']:
    module = __import__(module, globals(), locals(), fromlist=None, level=1)
    for name in dir(module):
        obj = getattr(module, name)
        if isinstance(obj, type) and issubclass(obj, Object):
            globals()[name] = obj



"""Add table for Abwasser.

Revision ID: 0593bb8c05b6
Revises: 9dadb9f9561f
Create Date: 2016-07-03 17:00:35.532851

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '0593bb8c05b6'
down_revision = '9dadb9f9561f'


def upgrade():
    op.create_table(
        'abwasser',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('parcel_id', sa.Integer(), nullable=False),
        sa.Column('value', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(['parcel_id'], [u'parcel.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    op.execute(
        "INSERT into abwasser (parcel_id, value) VALUES (85, 600000);")
    op.execute(
        "INSERT into abwasser (parcel_id, value) VALUES (86, 600000);")
    op.execute(
        "INSERT into abwasser (parcel_id, value) VALUES (116, 600000);")
    op.execute(
        "INSERT into abwasser (parcel_id, value) VALUES (153, 600000);")
    op.execute(
        "INSERT into abwasser (parcel_id, value) VALUES (119, 600000);")
    op.execute(
        "INSERT into abwasser (parcel_id, value) VALUES (120, 600000);")
    op.execute(
        "INSERT into abwasser (parcel_id, value) VALUES (177, 600000);")
    op.execute(
        "INSERT into abwasser (parcel_id, value) VALUES (175, 600000);")
    op.execute(
        "INSERT into abwasser (parcel_id, value) VALUES (128, 600000);")
    op.execute(
        "INSERT into abwasser (parcel_id, value) VALUES (159, 600000);")
    op.execute(
        "INSERT into abwasser (parcel_id, value) VALUES (134, 600000);")
    op.execute(
        "INSERT into abwasser (parcel_id, value) VALUES (165, 600000);")
    op.execute(
        "INSERT into abwasser (parcel_id, value) VALUES (147, 600000);")
    op.execute(
        "INSERT into abwasser (parcel_id, value) VALUES (172, 300000);")
    op.execute(
        "INSERT into abwasser (parcel_id, value) VALUES (174, 300000);")
    op.execute(
        "INSERT into abwasser (parcel_id, value) VALUES (99, 600000);")


def downgrade():
    op.drop_table('abwasser')



"""Add organization, phone and note to members/externalrecipients

Revision ID: 92cec0e8663e
Revises: e2cfe9da21cf
Create Date: 2020-10-22 08:11:26.996989

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '92cec0e8663e'
down_revision = 'e2cfe9da21cf'


def upgrade():
    op.add_column(
        'externalrecipient',
        sa.Column('organization', sa.String(length=100), nullable=True)
    )
    op.add_column(
        'externalrecipient',
        sa.Column('phone', sa.String(length=50), nullable=True)
    )
    op.add_column(
        'member',
        sa.Column('note', sa.Text(), nullable=True)
    )
    op.add_column(
        'member',
        sa.Column('organization', sa.String(length=100), nullable=True)
    )


def downgrade():
    op.drop_column('member', 'organization')
    op.drop_column('member', 'note')
    op.drop_column('externalrecipient', 'phone')
    op.drop_column('externalrecipient', 'organization')



# encoding=utf8
from sw.allotmentclub.conftest import import_members
from datetime import datetime
import pkg_resources
import transaction
import mock
import pytest


@pytest.mark.xfail
def test_MapView__call__1(browser, json_fixture):
    """It returns members and parcels."""
    import_members()
    transaction.commit()
    browser.login()
    url = json_fixture.url()
    browser.open('http://localhost{}'.format(url))
    json_fixture.assertEqual(browser.json['data'], 'data')
    json_fixture.assertEqual(browser.json['data'], 'map')
    json_fixture.assertEqual(browser.json['data'], 'map_data')


def test_MapDownloadView__call__1(browser, json_fixture):
    """Map can be downloaded."""
    expected = json_fixture.fixture
    browser.login()
    with mock.patch('sw.allotmentclub.browser.map.datetime') as dt:
        dt.now.return_value = datetime(2015, 2, 24, 8)
        browser.post(
            'http://localhost{}'.format(expected['url']),
            data=expected['data'],
            type=expected['type'])
    # assertFileEqual(browser.contents, 'test_map_print.pdf')
    with open(pkg_resources.resource_filename(
        'sw.allotmentclub.browser.tests', 'test_map_print.svg'
    ), 'r') as f:
        assert f.read() == browser.contents



# encoding=utf-8
from __future__ import unicode_literals
from ..conftest import import_bookings, Amount
import datetime
import pytest


def _getOne():
    from sw.allotmentclub import Booking
    return Booking.query().filter(
        Booking.purpose == (
            "Vereinsbeitrag für 2015 Parzelle 214 Czech / Schroeter")
    ).one()


def test_import_via_sparkassen_csv(database):
    from sw.allotmentclub import Booking, BankingAccount
    import_bookings()
    assert 1 == len(BankingAccount.query().all())
    assert 30 == len(Booking.query().all())


def test_imported_bookings_do_contain_iban_and_bic(database):
    import_bookings()
    assert 'DE79800537624440000377' == _getOne().iban
    assert 'NOLADE21HAL' == _getOne().bic


def test_imported_booking_data(database):
    import_bookings()
    booking = _getOne()
    assert '3440000167' == booking.banking_account.number
    assert '2015-01-19' == booking.booking_day.strftime('%Y-%m-%d')
    assert 'GUTSCHRIFT' == booking.booking_text
    assert u'Horst Schr\xf6ter u. Frau Ingr' == booking.recipient
    assert 600000 == booking.value


def test_booking_can_be_splitted(database):
    from sw.allotmentclub import Booking
    import_bookings()
    booking = _getOne()
    booking.split(259000)
    assert booking.is_splitted is True
    assert booking.splitted_from_id is None

    splitted = Booking.query().filter(
        Booking.splitted_from_id == booking.id).all()
    assert 2 == len(splitted)
    assert 259000 == splitted[0].value
    assert 341000 == splitted[1].value


def test_split_booking_failure_1(database):
    from sw.allotmentclub import Booking
    booking = Booking.create(value=10)
    assert booking.split(20) == 'Neuer Wert ist zu groß.'


def test_split_booking_failure_2(database):
    from sw.allotmentclub import Booking
    booking = Booking.create(value=-10)
    assert booking.split(-20) == 'Neuer Wert ist zu klein.'


def test_split_booking_failure_3(database):
    from sw.allotmentclub import Booking
    booking = Booking.create(value=10)
    assert booking.split(-20) == 'Neuer Wert muss positiv sein.'


def test_split_booking_failure_4(database):
    from sw.allotmentclub import Booking
    booking = Booking.create(value=-10)
    assert booking.split(20) == 'Neuer Wert muss negativ sein.'


def test_split_booking_failure_5(database):
    from sw.allotmentclub import Booking
    booking = Booking.create()
    assert booking.split(0) == 'Neuer Wert muss größer oder kleiner 0 sein.'


def test_imported_bookings_are_auto_assigned(database, member):
    from sw.allotmentclub import Member
    member = Member.query().one()
    member.iban = 'DE79800537624440000377'
    member.bic = 'NOLADE21HAL'
    import_bookings()
    booking = _getOne()
    assert 'Mittag' == booking.member.lastname


def test_import_sepa_raises_value_error_if_no_sammler(database, member):
    from ..account import import_transactions, BankingAccount, Booking
    account = BankingAccount.find_or_create(
        organization_id=1, number='3440000167')
    statements = [{
        'date': datetime.datetime.now(),
        'currency': 'EUR',
        'purpose': 'Energieabrechnung',
        'amount': Amount(2743.12, 'EUR'),
        'applicant_name': 'Verein',
        'posting_text': 'SAMMEL-LS-EINZUG',
        'applicant_iban': 'DE123456778899',
        'applicant_bin': 'NOLADE21HAL',
        'customer_reference': ('DATUM 14.06.2018, 14.28 UHR ANZAHL 9\n'
                               'PII7ebef1ebc3ec4a5185583851c8f7dbb2')
    }]
    assert 0 == Booking.query().count()
    with pytest.raises(
            ValueError,
            match=(r"Keine Sammler zu PmtInfId "
                   r"'PII7ebef1ebc3ec4a5185583851c8f7dbb2' gefunden.")):
        import_transactions(statements, account)
    assert 0 == Booking.query().count()


def test_import_sepa_raises_value_error_if_empty_sammler(database, member):
    from ..account import import_transactions, BankingAccount, SEPASammler
    from ..account import Booking
    account = BankingAccount.find_or_create(
        organization_id=1, number='3440000167')
    SEPASammler.create(pmtinfid='PII7ebef1ebc3ec4a5185583851c8f7dbb2',
                       booking_day=datetime.datetime.now(),
                       accounting_year=2018)
    statements = [{
        'date': datetime.datetime.now(),
        'currency': 'EUR',
        'purpose': 'Energieabrechnung',
        'amount': Amount(2743.12, 'EUR'),
        'applicant_name': 'Verein',
        'posting_text': 'SAMMEL-LS-EINZUG',
        'applicant_iban': 'DE123456778899',
        'applicant_bin': 'NOLADE21HAL',
        'customer_reference': ('DATUM 14.06.2018, 14.28 UHR ANZAHL 9\n'
                               'PII7ebef1ebc3ec4a5185583851c8f7dbb2')
    }]
    assert 0 == Booking.query().count()
    with pytest.raises(
            ValueError,
            match=(r"Keine Sammler-Einträge zu PmtInfId "
                   r"'PII7ebef1ebc3ec4a5185583851c8f7dbb2' gefunden.")):
        import_transactions(statements, account)
    assert 0 == Booking.query().count()


def test_import_sepa_creates_new_booking_entry_for_open_amount(database,
                                                               member):
    from ..account import import_transactions, SEPASammler, SEPASammlerEntry
    from ..account import BankingAccount, Booking
    account = BankingAccount.find_or_create(
        organization_id=1, number='3440000167')
    sammler = SEPASammler.create(
        pmtinfid='PII7ebef1ebc3ec4a5185583851c8f7dbb2',
        booking_day=datetime.datetime.now(),
        accounting_year=2018)
    SEPASammlerEntry.create(sepasammler=sammler, value=10000000)
    statements = [{
        'date': datetime.datetime.now(),
        'currency': 'EUR',
        'purpose': 'Energieabrechnung',
        'amount': Amount(2743.12, 'EUR'),
        'applicant_name': 'Verein',
        'posting_text': 'SAMMEL-LS-EINZUG',
        'applicant_iban': 'DE123456778899',
        'applicant_bin': 'NOLADE21HAL',
        'customer_reference': ('DATUM 14.06.2018, 14.28 UHR ANZAHL 9\n'
                               'PII7ebef1ebc3ec4a5185583851c8f7dbb2')
    }]
    assert 0 == Booking.query().count()
    import_transactions(statements, account)
    assert 3 == Booking.query().count()

    assert 27431200 == Booking.get(1).value
    assert 10000000 == Booking.get(2).value
    assert 17431200 == Booking.get(3).value

    assert 1 == Booking.get(2).splitted_from_id
    assert 1 == Booking.get(3).splitted_from_id

    assert 'SAMMEL-LS-EINZUG Verrechnet' == Booking.get(2).booking_text



# coding:utf8
from babel.numbers import format_currency
import datetime


def format_kwh(value):
    return '{} kWh'.format(value if value is not None else '--')


def format_eur(value, full=False):
    if value is None:
        return '---.-- €'
    value = value / 10000.0
    format = '#,##0.00 ¤'
    if full:
        format = '#,##0.0000 ¤'
    return format_currency(
        value, 'EUR', format, currency_digits=False, locale='de_DE')


def parse_date(value):
    if not value:
        return None
    if isinstance(value, datetime.date):
        return value
    if isinstance(value, datetime.datetime):
        return value
    fmts = [
        '%Y-%m-%dT%H:%M:%S%z',
        '%Y-%m-%dT%H:%M:%S',
        '%d.%m.%Y %H:%M',
        '%d.%m.%Y',
        '%Y-%m-%d %H:%M',
        '%Y-%m-%d',
    ]
    for fmt in fmts:
        try:
            return datetime.datetime.strptime(value, fmt)
        except (ValueError, TypeError):
            continue
    raise ValueError(f'Could not parse {value} into da datetime object')



"""re #3: Remove no longer used code.

Revision ID: 1dfd521b7ca0
Revises: d1ccee162c46
Create Date: 2016-07-26 12:45:07.547110

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '1dfd521b7ca0'
down_revision = 'd1ccee162c46'


def upgrade():
    op.drop_column('user', 'read_only')
    op.drop_column('user', 'is_revision')
    op.drop_column('user', 'can_modify')


def downgrade():
    op.add_column(
        'user',
        sa.Column('can_modify', sa.BOOLEAN(), server_default=sa.text(u'false'),
                  autoincrement=False, nullable=False))
    op.add_column(
        'user',
        sa.Column('is_revision', sa.BOOLEAN(),
                  server_default=sa.text(u'false'), autoincrement=False,
                  nullable=False))
    op.add_column(
        'user',
        sa.Column('read_only', sa.BOOLEAN(), server_default=sa.text(u'false'),
                  autoincrement=False, nullable=False))



"""re #12: Add column to save passive allotment

Revision ID: 3efd1e093b36
Revises: 53f3d16657d3
Create Date: 2021-01-02 16:19:38.218669

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '3efd1e093b36'
down_revision = '53f3d16657d3'


def upgrade():
    op.add_column(
        'member',
        sa.Column('passive_allotment_id', sa.Integer(), nullable=True)
    )
    op.create_foreign_key(
        'passive_allotment_id_fkey', 'member', 'allotment',
        ['passive_allotment_id'], ['id']
    )


def downgrade():
    op.drop_constraint(
        'passive_allotment_id_fkey', 'member', type_='foreignkey'
    )
    op.drop_column('member', 'passive_allotment_id')



"""re #10: Add event model

Revision ID: 2952c5487114
Revises: c4dbe65a7810
Create Date: 2021-01-06 12:51:42.034633

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '2952c5487114'
down_revision = 'c4dbe65a7810'


def upgrade():
    op.create_table(
        'event',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('organization_id', sa.Integer(), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=True),
        sa.Column('category', sa.String(), nullable=True),
        sa.Column('title', sa.String(), nullable=True),
        sa.Column('description', sa.String(), nullable=True),
        sa.Column('start', sa.DateTime(), nullable=True),
        sa.Column('end', sa.DateTime(), nullable=True),
        sa.Column('allday', sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(['user_id'], ['user.id'], ),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('event')



"""Add email to members.

Revision ID: 2cb8d240d5ac
Revises: 4e4d0ae80e7a
Create Date: 2015-03-02 15:07:38.712318

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '2cb8d240d5ac'
down_revision = '4e4d0ae80e7a'


def upgrade():
    op.add_column(
        'member', sa.Column('email', sa.String(length=100), nullable=True))
    op.execute("""UPDATE member SET email = 'sebastian.wehrmann@icloud.com'
                  WHERE id = 76;""")
    op.execute(
        "UPDATE member SET email = 'stephan.leuna@t-online.de' WHERE id = 85;")
    op.execute(
        "UPDATE member SET email = 'andre_atb@msn.com' WHERE id = 86;")
    op.execute(
        "UPDATE member SET email = 'kalovo39@gmx.de' WHERE id = 89;")
    op.execute(
        "UPDATE member SET email = 'smjala@aim.com' WHERE id = 90;")
    op.execute(
        "UPDATE member SET email = 'olaf.kuehn@gmx.net' WHERE id = 91;")
    op.execute(
        "UPDATE member SET email = 's.kern@ivh.de' WHERE id = 100;")
    op.execute(
        "UPDATE member SET email = 'amielke@primacom.net' WHERE id = 101;")
    op.execute("""UPDATE member SET email = 'j.sonntag@leipziger-messe.de'
                  WHERE id = 103;""")
    op.execute(
        "UPDATE member SET email = 'harry.noglik@t-online.de' WHERE id = 105;")
    op.execute("""UPDATE member SET email = 'lochau.hartmann@freenet.de'
                  WHERE id = 118;""")
    op.execute(
        "UPDATE member SET email = 'morki1@hotmail.de' WHERE id = 123;")
    op.execute(
        "UPDATE member SET email = 'boehme.wsf@t-online.de' WHERE id = 126;")
    op.execute(
        "UPDATE member SET email = 'juergen.liehm@web.de' WHERE id = 127;")
    op.execute("""UPDATE member SET email = 'roesler@informatik.uni-halle.de'
                  WHERE id = 129;""")
    op.execute(
        "UPDATE member SET email = 'super_bingo@freenet.de' WHERE id = 133;")
    op.execute(
        "UPDATE member SET email = 'spiess.dietmar@web.de' WHERE id = 134;")
    op.execute(
        "UPDATE member SET email = 'eberhard.kietz@online.de' WHERE id = 135;")
    op.execute("""UPDATE member SET email = 'peter.reinhold@superkabel.de'
                  WHERE id = 141;""")
    op.execute(
        "UPDATE member SET email = 'tee.mueller@t-online.de' WHERE id = 142;")
    op.execute(
        "UPDATE member SET email = 'b.e.bochi@gmx.de' WHERE id = 149;")
    op.execute(
        "UPDATE member SET email = 'ominette@gmx.de' WHERE id = 150;")
    op.execute(
        "UPDATE member SET email = 'dr.kobelt@gmx.de' WHERE id = 156;")
    op.execute(
        "UPDATE member SET email = 'r.lichtenfeld@freenet.de' WHERE id = 157;")
    op.execute(
        "UPDATE member SET email = 'basteu@t-online.de' WHERE id = 158;")
    op.execute(
        "UPDATE member SET email = 'angela_maxis@arcor.de' WHERE id = 164;")
    op.execute(
        "UPDATE member SET email = 'gerd-wmi@t-online.de' WHERE id = 165;")
    op.execute(
        "UPDATE member SET email = 'uwe.reinicke@epc.com' WHERE id = 167;")
    op.execute(
        "UPDATE member SET email = 'fischereddager@aol.com' WHERE id = 171;")
    op.execute(
        "UPDATE member SET email = 'beritschmidt@arcor.de' WHERE id = 173;")
    op.execute(
        "UPDATE member SET email = 'klaus-hertwig@online.de' WHERE id = 174;")
    op.execute(
        "UPDATE member SET email = 'GuenterTillack@aim.com' WHERE id = 185;")
    op.execute(
        "UPDATE member SET email = 'schloeffel_h@t-online.de' WHERE id = 186;")


def downgrade():
    op.drop_column('member', 'email')



"""Add new accounts

Revision ID: 48303c2cf2eb
Revises: e927a9a287ad
Create Date: 2023-06-19 07:53:42.089315

"""
from alembic import op
import base64
import pkg_resources

# revision identifiers, used by Alembic.
revision = '48303c2cf2eb'
down_revision = 'e927a9a287ad'


def upgrade():
    signature = pkg_resources.resource_stream(
        'sw.allotmentclub.signatures', 'rw.png').read()
    signature = 'data:application/png;base64,{}'.format(
        base64.b64encode(signature).decode('utf-8'))
    op.execute("""
INSERT into public.user
    (id, username, password, unrestricted_access, vorname, nachname, anrede,
     handynummer, ort, position, signature, email, organization_id)
VALUES
    (6, 'rw', '{pw}', true, 'Ronny', 'Weber', 'Herr', '0173 2713035',
     'Leipzig', 'Stellvertretender Vorsitzender', '{signature}',
     'ronnyweber11@aol.com', 1)
""".format(pw='$2a$12$23Lz6/ozpjpMd0NMxGWfa.dcGHG6/6gHptb8c7U5TLDXXElW6RZ96',
           signature=signature))

    signature = pkg_resources.resource_stream(
        'sw.allotmentclub.signatures', 'ms.png').read()
    signature = 'data:application/png;base64,{}'.format(
        base64.b64encode(signature).decode('utf-8'))
    op.execute("""
INSERT into public.user
    (id, username, password, unrestricted_access, vorname, nachname, anrede,
     handynummer, ort, position, signature, email, organization_id)
VALUES
    (8, 'ms', '{pw}', false, 'Matthias', 'Schulz', 'Herr', '0152 26346780',
     'Bad Lauchstädt', 'Schatzmeister', '{signature}',
     'matthias.schulz82@gmail.com', 1)
""".format(pw='$2a$12$nFSJLH7MVmUPqw4TZLz9uO2eAhvyEGOceLiT7BdaqMkNltRcY7x3e',
           signature=signature))

    op.execute("""
UPDATE public.user SET is_locked = 'Zurückgetreten' WHERE username = 'hs'""")
    op.execute("""
UPDATE public.user SET is_locked = 'Zurückgetreten' WHERE username = 'ar'""")
    op.execute("""
UPDATE public.user SET position = 'Vorsitzende', unrestricted_access = true
WHERE username = 'ah'""")
    op.execute("UPDATE accessauthority SET user_id = 8 WHERE user_id = 41")


def downgrade():
    op.execute("UPDATE public.user SET is_locked = null WHERE username = 'hs'")
    op.execute("UPDATE public.user SET is_locked = null WHERE username = 'ar'")
    op.execute("""\
UPDATE public.user SET position = 'Schatzmeister',
unrestricted_access = false WHERE username = 'ah'""")
    op.execute("UPDATE accessauthority SET user_id = 41 WHERE user_id = 8")
    op.execute("DELETE FROM log WHERE user_id in (6, 8);")
    op.execute("DELETE FROM public.user WHERE username = 'ms';")
    op.execute("DELETE FROM public.user WHERE username = 'rw';")



"""Increase mimetype space for depot

Revision ID: 74c2dd192dae
Revises: 2a39970710de
Create Date: 2018-05-28 10:10:39.194098

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '74c2dd192dae'
down_revision = '2a39970710de'


def upgrade():
    op.alter_column('depot', 'mimetype', type_=sa.VARCHAR(100))


def downgrade():
    op.alter_column('depot', 'mimetype', type_=sa.VARCHAR(30))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import datetime
import transaction


def setUp():
    from sw.allotmentclub import Bulletin, User
    user = User.create(username='hans')
    Bulletin.create(
        date=datetime.datetime(2015, 3, 7, 10),
        subject='Ruhezeiten im Verein',
        user=user,
        content="""
### Mittagsruhe

- täglich 13.00 - 15.00 Uhr (ganzjährig)

### Nachtruhe

- Montag bis Freitag:      22.00 - 08.00 Uhr
- Samstag &amp; Sonntag:    24.00 - 08.00 Uhr

### Genereller Baustop

- jährlich vom 15. Juni bis 31. August.""")
    transaction.commit()


def test_BulletinListView_1(browser, json_fixture):
    """It displays list of bulletins."""
    url = json_fixture.url()
    setUp()
    browser.login()
    browser.open('http://localhost{}'.format(url))
    json_fixture.assertEqual(browser.json, 'data')


def test_BulletinAddView_1(browser, json_fixture):
    """It can add new bulletins via JSON."""
    url = json_fixture.url()
    browser.login()
    browser.post('http://localhost{}'.format(url), data='')
    assert 'success' == browser.json['status']
    json_fixture.assertEqual(browser.json, 'data')


def test_BulletinEditView_1(browser, json_fixture):
    """It can update bulletins from JSON."""
    from sw.allotmentclub import Bulletin
    setUp()
    assert 'Vereinsfest findet statt' != Bulletin.query().one().subject
    browser.login()
    json_fixture.ajax(browser)
    assert 'success' == browser.json['status']
    assert 'Vereinsfest findet statt' == Bulletin.query().one().subject


def test_BulletinDeleteView_1(browser, json_fixture):
    from sw.allotmentclub import Bulletin
    setUp()
    assert 1 == len(Bulletin.query().all())
    browser.login()
    json_fixture.ajax(browser)
    assert 'success' == browser.json['status']
    assert 0 == len(Bulletin.query().all())


def test_bulletins_can_be_printed_as_pdf(browser):
    setUp()
    browser.login()
    browser.open('http://localhost/bulletins/1/print')
    assert browser.contents.decode('latin1').startswith('%PDF-')



# encoding=utf8
from __future__ import unicode_literals
from ..log import user_data_log, log_with_user
from .base import date, get_selected_year, value_to_int, boolean, date_time
from .base import format_eur_with_color, format_eur, format_date, to_string
from .base import string_agg, parse_date
from .mail import append_pdf, render_pdf, format_markdown
from PyPDF2 import PdfFileWriter, PdfFileReader
from io import BytesIO
from pyramid.response import FileIter
from pyramid.view import view_config
from sqlalchemy.sql import func
from sw.allotmentclub import Booking, Member, Allotment, GrundsteuerB, Abwasser
from sw.allotmentclub import BookingKind, SEPASammler, SEPASammlerEntry
from sw.allotmentclub import Budget
import collections
import datetime
import pybars
import re
import sqlalchemy
import sw.allotmentclub.browser.base


def get_booking_future(days=90):
    return datetime.datetime.now() + datetime.timedelta(days=days)


def AccountDetailFactory(request):
    account_id, account_type = request.matchdict['id'].split('-')
    context = getattr(sw.allotmentclub.account,
                      account_type).get(account_id)
    context.__acl__ = sw.allotmentclub.browser.auth.DEFAULT_ACL
    return context


class Query(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Datum': date,
        'Betrag': format_eur_with_color,
    }

    css_classes = {
        'Betrag': 'right nowrap',
        'Datum': 'nowrap',
        'Zugeordnet': 'nowrap',
    }

    data_class = {
        'Buchungsinfo': 'expand'
    }
    data_hide = {
        'Datum': 'phone',
        'Betrag': 'phone',
        'Zugeordnet': 'phone,tablet',
    }

    def select(self):
        from sqlalchemy.sql.expression import false
        return (
            self.db.query(
                Booking.id.label('#'),
                Booking.booking_day.label('Datum'),
                (to_string(Booking.booking_text).concat('<br />')
                 .concat(to_string(Booking.purpose)).concat('<br />')
                 .concat(to_string(Booking.recipient)).label('Buchungsinfo')),
                Booking.value.label('Betrag'),
                (to_string(Member.lastname).concat(' (')
                 .concat(func.coalesce(string_agg(Allotment.number), 'n/a'))
                 .concat(')').label('Zugeordnet')),
                BookingKind.shorttitle.label('Typ'),
            )
            .select_from(Booking)
            .outerjoin(Member)
            .outerjoin(Allotment, Allotment.member_id == Member.id)
            .outerjoin(BookingKind)
            .group_by(Booking.id, Member.lastname, BookingKind.shorttitle)
            .filter(Booking.banking_account == self.context)
            .filter(Booking.is_splitted == false())
            .filter(Booking.accounting_year == get_selected_year()))


@view_config(route_name='booking_list', renderer='json',
             permission='view')
class BookingListView(sw.allotmentclub.browser.base.TableView):

    query_class = Query
    default_order_by = 'booking_day'
    year_selection = True
    available_actions = [
        dict(url='map_booking', btn_class='btn-success',
             icon='fa fa-hand-o-right', title='Zuordnen'),
        dict(url='split_booking', btn_class='btn-success',
             icon='fa fa-scissors', title='Aufteilen')
    ]


@view_config(route_name='split_booking', renderer='json',
             permission='view')
class SplitBookingView(sw.allotmentclub.browser.base.EditJSFormView):

    title = 'Buchung aufteilen'

    @property
    def load_options(self):
        return {
            'new_value': {
                'label': 'Teilen bei Betrag',
                'required': True
            }
        }

    @property
    def load_data(self):
        fields = [
            ('new_value', '')
        ]
        return collections.OrderedDict(fields)

    def fail(self, message):
        self.result['status'] = 'error'
        self.result['msg'] = message

    def update(self):
        try:
            json = self.request.json
        except ValueError:
            json = None
        if json:
            new_value = value_to_int(json['new_value'])
            status = self.context.split(new_value)
            if isinstance(status, str):
                self.fail(status)
                return
            log_with_user(
                user_data_log.info, self.request.user,
                'Buchung %s bei %s aufgeteilt.',
                self.context.id, format_eur(new_value))
            self.result = {'status': 'success'}
        else:
            self.result = self.get_result()


@view_config(route_name='map_booking', renderer='json',
             permission='view')
class MapBookingView(sw.allotmentclub.browser.base.EditJSFormView):

    title = 'Buchung einem Mitglied zuweisen'

    @property
    def load_options(self):
        result = {
            'member_id': {
                'source': self.member_source,
                'label': 'Mitglied',
                'css_class': 'chosen',
                'required': False
            },
            'kind_id': {
                'source': self.booking_kind_source,
                'label': 'Typ',
                'css_class': 'chosen',
                'required': False
            },
            'accounting_year': {
                'source': self.accounting_year_source,
                'label': 'Jahr',
                'css_class': 'chosen',
                'required': True
            }
        }
        return result

    @property
    def load_data(self):
        fields = [
            ('member_id', self.context.member_id),
            ('kind_id', self.context.kind_id),
            ('accounting_year', self.context.accounting_year)]
        return collections.OrderedDict(fields)


def get_balance(value, request=None):
    from sqlalchemy.sql.expression import true, false, case
    import zope.component
    import risclog.sqlalchemy
    member = Member.get(value)
    db = zope.component.getUtility(risclog.sqlalchemy.interfaces.IDatabase)
    bookings = (
        db.query(func.sum(Booking.value))
        .join(Member)
        .filter(Booking.member == member)
        .filter(Booking.is_splitted == false())
        .filter(Booking.booking_day <= get_booking_future())
        .filter(Booking.accounting_year == get_selected_year()))
    sepa = (
        db.query(func.sum(
            case([
                (
                    SEPASammler.is_ueberweisung == true(),
                    0 - SEPASammlerEntry.value
                )],
                else_=SEPASammlerEntry.value,
            ).label('Betrag'),
        ))
        .join(Member)
        .join(SEPASammler)
        .filter(SEPASammlerEntry.member == member)
        .filter(SEPASammler.booking_day <= get_booking_future())
        .filter(SEPASammler.accounting_year == get_selected_year()))
    booking_sum = bookings.one()[0]
    sepa_sum = sepa.one()[0]
    booking_sum = 0 if booking_sum is None else booking_sum
    sepa_sum = 0 if sepa_sum is None else sepa_sum
    return format_eur_with_color(booking_sum + sepa_sum)


class MemberAccountQuery(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Konto': get_balance,
    }

    css_classes = {
        'Konto': 'right nowrap',
    }

    data_class = {
        'Mitglied': 'expand'
    }
    data_hide = {
        'Konto': 'phone',
    }

    def select(self):
        return (
            self.db.query(
                Member.id.label('#'),
                (to_string(Member.lastname).concat(', ')
                 .concat(to_string(Member.firstname)).label('Mitglied')),
                Member.id.label('Konto')
            )
            .select_from(Member)
            .filter(sqlalchemy.or_(
                Member.leaving_year.is_(None),
                Member.leaving_year >= get_selected_year())))


@view_config(route_name='member_account_list', renderer='json',
             permission='view')
class MemberAccountListView(sw.allotmentclub.browser.base.TableView):

    query_class = MemberAccountQuery
    default_order_by = 'member.id'
    year_selection = True
    available_actions = [
        dict(url='member_account_detail_list', btn_class='btn-success',
             icon='fa fa-list', title='Buchungen'),
    ]


class MemberAccountDetailQuery(sw.allotmentclub.browser.base.Query):

    disable_global_organization_filter = True

    formatters = {
        'Betrag': format_eur_with_color,
        'Datum': date,
        'RS': boolean
    }

    css_classes = {
        'Betrag': 'right nowrap',
    }

    data_class = {
        'Datum': 'expand'
    }
    data_hide = {
        'Betrag': 'phone',
        'Typ': 'phone, tablet',
    }

    def select(self):
        from sqlalchemy.sql.expression import true, false, case
        bookings = (
            self.db.query(
                to_string(Booking.id).concat('-Booking').label('#'),
                Booking.booking_day.label('Datum'),
                BookingKind.title.label('Typ'),
                Booking.value.label('Betrag'),
                Booking.ignore_in_reporting.label('RS')
            )
            .select_from(Booking)
            .outerjoin(BookingKind)
            .join(Member)
            .filter(Booking.is_splitted == false())
            .filter(Booking.accounting_year == get_selected_year())
            .filter(Booking.booking_day <= get_booking_future())
            .filter(Booking.organization_id == self.user.organization_id)
            .filter(Booking.member == self.context))
        sepa = (
            self.db.query(
                (to_string(SEPASammlerEntry.id)
                 .concat('-SEPASammlerEntry').label('#')),
                SEPASammler.booking_day.label('Datum'),
                BookingKind.title.label('Typ'),
                case([
                    (
                        SEPASammler.is_ueberweisung == true(),
                        0 - SEPASammlerEntry.value
                    )],
                    else_=SEPASammlerEntry.value,
                ).label('Betrag'),
                SEPASammlerEntry.ignore_in_reporting.label('RS')
            )
            .select_from(SEPASammler)
            .join(SEPASammlerEntry)
            .outerjoin(BookingKind)
            .join(Member)
            .filter(SEPASammler.accounting_year == get_selected_year())
            .filter(SEPASammler.booking_day <= get_booking_future())
            .filter(SEPASammler.organization_id == self.user.organization_id)
            .filter(SEPASammlerEntry.member == self.context))
        return bookings.union(sepa)


@view_config(route_name='member_account_detail_list', renderer='json',
             permission='view')
class MemberAccountDetailListView(sw.allotmentclub.browser.base.TableView):

    query_class = MemberAccountDetailQuery
    year_selection = True
    available_actions = [
        dict(url='member_account_detail_switch_ir',
             btn_class='btn-danger',
             icon='fa fa-check-square-o ', title='Reporting-Status ändern'),
    ]


@view_config(route_name='member_account_detail_switch_ir', renderer='json',
             permission='view')
class MemberAccountDetailSwitchRSView(sw.allotmentclub.browser.base.View):

    def update(self):
        self.context.ignore_in_reporting = not self.context.ignore_in_reporting
        self.result = {'status': 'success',
                       'message': 'Reporting-Status geändert.'}


def get_budget_raw(value, request=None, year=None):
    year = year if year else get_selected_year()
    kind = BookingKind.get(value)
    budget = (
        Budget.query()
        .filter(Budget.booking_kind == kind)
        .filter(Budget.accounting_year == year)
        .one_or_none()
    )
    if budget:
        return [budget]
    return []


def get_budget_sum(value, request=None, year=None):
    year = year if year else get_selected_year()
    kind = BookingKind.get(value)
    if kind.shorttitle == 'ENAB':
        kind = BookingKind.query().filter_by(shorttitle='ENAN').one()
        query = (
            Booking.query()
            .filter(Booking.kind == kind)
            .filter(Booking.accounting_year == year)
        )
        return 0 - sum(b.value for b in query)
    return sum(b.value for b in get_budget_raw(value, request, year))


def get_budget(value, request=None, year=None):
    return format_eur_with_color(get_budget_sum(value, request, year))


def get_incoming_raw(value, request=None, year=None):
    from sqlalchemy.sql.expression import false
    year = year if year else get_selected_year()
    kind = BookingKind.get(value)
    query = (
        Booking.query()
        .join(Member)
        .filter(Booking.kind == kind)
        .filter(Booking.accounting_year == year)
        .filter(Booking.ignore_in_reporting == false())
        .filter(Booking.is_splitted == false())
        .filter(Booking.banking_account_id == 1)
        .filter(Booking.booking_text != 'SAMMEL-LS-EINZUG')
        .filter(~Booking.purpose.like('%ANZAHL 00000%'))
    )
    sammler = (
        SEPASammlerEntry.query()
        .join(SEPASammler)
        .join(Member)
        .filter(SEPASammler.kind == kind)
        .filter(SEPASammlerEntry.ignore_in_reporting == false())
        .filter(SEPASammler.accounting_year == year)
    )
    result = query.all() + sammler.all()
    if kind.shorttitle == 'ENAB':
        result += (
            get_incoming_raw(
                BookingKind.query().filter_by(shorttitle='ENA1').one().id,
                request, year
            ) +
            get_incoming_raw(
                BookingKind.query().filter_by(shorttitle='ENA2').one().id,
                request, year
            )
        )
    return result


def get_incoming_sum(value, request=None, year=None):
    result = 0
    for item in get_incoming_raw(value, request, year):
        value = item.value
        if hasattr(item, 'sepasammler'):
            if item.sepasammler.is_ueberweisung:
                value = 0 - value
        result += value
    return result


def get_incoming(value, request=None, year=None):
    return format_eur_with_color(get_incoming_sum(value, request, year))


def get_incoming_pre_year(value, request=None):
    return get_incoming(value, request, get_selected_year()-1)


def get_sum(value, request=None):
    budget = get_budget_sum(value, request)
    actual = get_incoming_sum(value, request)

    return format_eur_with_color(actual - budget)


class BankingAccountQuery(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Vorjahr': get_incoming_pre_year,
        'Budget': get_budget,
        'Ist': get_incoming,
        'Differenz': get_sum,
    }

    css_classes = {
        'Vorjahr': 'right nowrap',
        'Budget': 'right nowrap',
        'Ist': 'right nowrap',
        'Differenz': 'right nowrap',
    }

    data_class = {
        'Kategorie': 'expand'
    }
    data_hide = {
        'Vorjahr': 'phone',
        'Budget': 'phone',
        'Ist': 'phone',
        'Differenz': 'phone,tablet'
    }

    def select(self):
        return (
            self.db.query(
                BookingKind.id.label('#'),
                BookingKind.title.label('Kategorie'),
                BookingKind.id.label('Vorjahr'),
                BookingKind.id.label('Budget'),
                BookingKind.id.label('Ist'),
                BookingKind.id.label('Differenz')
            )
            .select_from(BookingKind)
            .filter(~BookingKind.shorttitle.in_(('ENA1', 'ENA2', 'ENAN')))
        )


@view_config(route_name='banking_account_list', renderer='json',
             permission='view')
class BankingAccountListView(sw.allotmentclub.browser.base.TableView):

    query_class = BankingAccountQuery
    year_selection = True
    available_actions = [
        dict(url='banking_account_list_detail', btn_class='btn-success',
             icon='fa fa-list', title='Positionen'),
        dict(url='banking_account_list_report', btn_class='btn-success',
             icon='fa fa-print', title='Report')
    ]


@view_config(route_name='banking_account_list_detail', renderer='json',
             permission='view')
class BankingAccountListDetailView(sw.allotmentclub.browser.base.TableView):

    def __call__(self):
        incoming = get_incoming_raw(self.context.id, self.request)
        columns = [
            {'name': name, 'css_class': 'hide' if id == 0 else ''}
            for id, name in enumerate(
                ['#', 'Betrag', 'Mitglied', 'Buchungstext', 'Datum'])]
        result = []
        for obj in incoming:
            value = obj.value
            if not hasattr(obj, 'booking_text'):
                if obj.sepasammler.is_ueberweisung:
                    type_ = 'Sammel-Überweisung'
                    value = 0 - value
                else:
                    type_ = 'Sammel-Lastschrift'
                text = (
                    f'{type_} {obj.sepasammler.kind.title} '
                    f'{obj.sepasammler.pmtinfid}'
                )
            else:
                text = obj.purpose
            new_line = [
                {'value': obj.id, 'css_class': 'hide'},
                {'value': format_eur_with_color(value, self.request)},
                {'value': '{}, {} ({})'.format(
                    obj.member.lastname, obj.member.firstname,
                    (obj.member.allotments[0].number
                        if obj.member.allotments else ''))},
                {'value': text},
                {'value': format_date(
                    obj.booking_day if hasattr(obj, 'booking_day')
                    else obj.sepasammler.booking_day)},
            ]
            result.append(new_line)
        data = {'data': result,
                'header': columns,
                'actions': [],
                'records': len(result)}
        return {'status': 'success', 'data': data}


@view_config(route_name='banking_account_list_report', renderer='json',
             permission='view')
class BankingAccountListReportView(sw.allotmentclub.browser.base.View):

    hbs_page = """
<h3>{{title}}</h3>
<table style="border: 1px solid black;">
  <tr style="font-style: normal; font-size: 8pt;">
    <th style="width: 10%; {{{header_style}}}">Betrag</th>
    <th style="width: 23%; {{{header_style}}}">Mitglied</th>
    <th style="width: 58%; {{{header_style}}}">Buchungstext</th>
    <th style="width: 9%; {{{header_style}}}">Datum</th>
  </tr>
  {{#each content}}
    <tr>
      {{#each this}}
      <td style="border: 1px solid black;
                 padding: 2px 2px -1px 2px;
                 {{#unless @index}} text-align: right; {{/unless}}
                 font-style: normal;
                 font-size: 8pt;">{{{this}}}</td>
      {{/each}}
    </tr>
  {{/each}}
  <tr style="font-style: normal; font-size: 8pt;">
    <th style="width: 10%; {{{header_style}}}">SUMME</th>
    <th colspan="3" style="{{{header_style}}}">{{summe}}</th>
  </tr>
</table>
"""

    def update(self):
        output = PdfFileWriter()
        compiler = pybars.Compiler()
        compiled_page = compiler.compile(self.hbs_page)
        cleanr = re.compile('<.*?>')
        year = get_selected_year()

        for kind in BookingKind.query():
            data = BankingAccountListDetailView(kind, self.request)()['data']
            table_content = []
            for item in data['data']:
                table_content.append(
                    [(re.sub(cleanr, '', d['value'])
                      if d['value'] else '&nbsp;') for d in item[1:5]])
            table_content = sorted(
                table_content,
                key=lambda x: (x[1], x[0], x[2]))  # sort by name
            sum_ = sum(float(
                i[0].replace(' €', '').replace('.', '').replace(',', '.'))
                for i in table_content)
            sum_ = int(sum_ * 10000)

            body = format_markdown(compiled_page(dict(
                title='{} {}'.format(kind.title, year),
                content=table_content,
                summe=format_eur(sum_),
                header_style=("border: 1px solid black; "
                              "padding: 2px 2px -1px 2px;"))))
            output = self.add_page(body, output)
        result = BytesIO()
        output.write(result)
        result.seek(0)
        response = self.request.response
        response.set_cookie('fileDownload', value='true')
        response.content_type = 'application/pdf'
        response.content_disposition = (
            f'attachment; filename=Report {year}.pdf'
        )
        response.app_iter = FileIter(result)
        self.result = response

    def add_page(self, body, output):
        pdf = render_pdf(None, None, body, None, date=None, force_from=False)
        pdf = PdfFileReader(pdf, strict=False)
        append_pdf(pdf, output)
        return output


class SEPASammlerQuery(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Datum': date,
    }
    data_class = {
        'Art': 'expand'
    }
    data_hide = {
        'Sparkassen-ID': 'phone,tablet',
        'Datum': 'phone,tablet',
    }

    def select(self):
        from sqlalchemy.sql.expression import case, literal, true
        return (
            self.db.query(
                SEPASammler.id.label('#'),
                BookingKind.title.label('Art'),
                SEPASammler.booking_day.label('Datum'),
                SEPASammler.pmtinfid.label('Sparkassen-ID'),
                case(
                    [
                        (
                            SEPASammler.is_ueberweisung == true(),
                            literal('Sammelüberweisung')
                        ),
                    ],
                    else_=literal('Sammellastschrift')
                ).label('Typ'),
            )
            .select_from(SEPASammler)
            .outerjoin(BookingKind)
            .filter(SEPASammler.accounting_year == get_selected_year()))


@view_config(route_name='sepa_sammler_list', renderer='json',
             permission='view')
class SEPASammlerListView(sw.allotmentclub.browser.base.TableView):

    query_class = SEPASammlerQuery
    year_selection = True
    start_year = 2015
    available_actions = [
        dict(url='sepa_sammler_add', btn_class='btn-success',
             icon='fa fa-plus', title='Neue Sammellastschrift'),
        dict(url='sepa_ueberweisung_add', btn_class='btn-success',
             icon='fa fa-plus', title='Neue Sammelüberweisung'),
        dict(url='sepa_sammler_edit', btn_class='btn-success',
             icon='fa fa-pencil', title='Bearbeiten'),
        dict(url='sepa_sammler_entry_list', btn_class='btn-success',
             icon='fa fa-list', title='Positionen'),
        dict(url='sepa_sammler_update', btn_class='btn-success',
             icon='fa fa-list', title='autom. Befüllen'),
        dict(url='sepa_sammler_export', btn_class='btn-success',
             icon='fa fa-bank', title='Export Sparkasse'),
    ]


@view_config(route_name='sepa_sammler_update', renderer='json',
             permission='view')
class SEPASammlerUpdateView(sw.allotmentclub.browser.base.View):

    def advancepay(self):
        values = (
            Booking.query()
            .filter(Booking.accounting_year == self.context.accounting_year)
            .filter(Booking.kind == self.context.kind).all())
        for value in values:
            if value.value < 0:
                if value.member and value.member.direct_debit:
                    SEPASammlerEntry().find_or_create(
                        sepasammler=self.context,
                        value=0 - value.value,
                        member=value.member)

    def energieabrechnung(self):
        values = (
            Booking.query()
            .filter(Booking.accounting_year == self.context.accounting_year)
            .filter(Booking.kind == self.context.kind)).all()
        if self.context.is_ueberweisung:
            values = [v for v in values if v.value > 0]
        else:
            values = [v for v in values if v.value <= 0]
        for value in values:
            if value.member and value.member.direct_debit:
                SEPASammlerEntry().find_or_create(
                    sepasammler=self.context,
                    value=(
                        value.value
                        if self.context.is_ueberweisung
                        else 0 - value.value
                    ),
                    member=value.member)

    def grundsteuerb(self):
        values = GrundsteuerB.query().all()
        for value in values:
            member = value.parcel.allotment.member
            if member.direct_debit:
                SEPASammlerEntry().find_or_create(
                    sepasammler=self.context,
                    value=value.value,
                    member=value.parcel.allotment.member)

    def abwasser(self):
        values = Abwasser.query().all()
        for value in values:
            member = value.parcel.allotment.member
            if member.direct_debit:
                SEPASammlerEntry().find_or_create(
                    sepasammler=self.context,
                    value=value.value,
                    member=member)

    @property
    def active_members(self):
        return super(SEPASammlerUpdateView, self).active_members.filter(
            Member.direct_debit.is_(True)
        )

    def arbeitsstunden(self):
        for member in self.active_members.all():
            bookings = (
                Booking.query()
                .filter(
                    Booking.accounting_year == self.context.accounting_year)
                .filter(Booking.kind == self.context.kind)
                .filter(Booking.member == member).all())
            value = 0
            for booking in bookings:
                value -= booking.value
            if value:
                SEPASammlerEntry().find_or_create(
                    sepasammler=self.context,
                    value=value,
                    member=member)

    def mitgliedsbeitrag(self):
        for booking in (
            Booking.query()
            .filter(Booking.accounting_year == self.context.accounting_year)
            .filter(Booking.kind == self.context.kind)
        ):
            if booking.member.direct_debit:
                SEPASammlerEntry().find_or_create(
                    sepasammler=self.context,
                    value=0 - booking.value,
                    member=booking.member)

    def update(self):
        if not self.context.kind:
            self.result = {'status': 'error',
                           'message': 'Art muss gesetzt sein.'}
            return
        for entry in (
                SEPASammlerEntry.query()
                .filter(SEPASammlerEntry.sepasammler == self.context).all()):
            entry.delete()
        if self.context.kind.shorttitle in ('ENA1', 'ENA2'):
            self.advancepay()
        elif self.context.kind.shorttitle == 'ENAB':
            self.energieabrechnung()
        elif self.context.kind.shorttitle == 'GRUB':
            self.grundsteuerb()
        elif self.context.kind.shorttitle == 'ABWA':
            self.abwasser()
        elif self.context.kind.shorttitle == 'NARB':
            self.arbeitsstunden()
        elif self.context.kind.shorttitle == 'BEIT':
            self.mitgliedsbeitrag()
        else:
            self.result = {
                'status': 'error',
                'message': '{} nicht unterstützt'.format(
                    self.context.kind.title)
            }
            return
        self.result = {
            'status': 'success',
            'message': 'SEPASammler {} berechnet.'.format(
                self.context.kind.title)}


@view_config(route_name='sepa_sammler_edit', renderer='json',
             permission='view')
class SEPASammlerEditView(sw.allotmentclub.browser.base.EditJSFormView):

    title = 'SEPA Sammler bearbeiten'

    @property
    def load_options(self):
        return {
            'kind_id': {
                'label': 'Art',
                'css_class': 'chosen',
                'required': True,
                'source': self.booking_kind_source,
            },
            'booking_day': {
                'label': 'Datum',
                'css_class': 'datepicker',
            },
            'pmtinfid': {
                'label': 'Sparkassen-ID',
            }
        }

    @property
    def booking_kind_source(self):
        result = super(SEPASammlerEditView, self).booking_kind_source
        if self.context.is_ueberweisung:
            return [r for r in result if r['title'] == 'Energieabrechnung']
        return result

    @property
    def load_data(self):
        fields = [
            ('kind_id', self.context.kind_id),
            ('booking_day', date_time(self.context.booking_day)),
            ('pmtinfid', self.context.pmtinfid),
        ]
        return collections.OrderedDict(fields)

    def save(self, key, value):
        if key == 'booking_day' and value:
            value = parse_date(value).date()
        return super(SEPASammlerEditView, self).save(key, value)

    def update(self):
        super(SEPASammlerEditView, self).update()
        if self.context.booking_day:
            if isinstance(self.context.booking_day, datetime.date):
                self.context.accounting_year = self.context.booking_day.year
            else:
                self.context.accounting_year = (
                    parse_date(self.context.booking_day).year)


@view_config(route_name='sepa_sammler_add', renderer='json',
             permission='view')
class SEPASammlerAddView(SEPASammlerEditView):

    def __init__(self, context, request):
        context = SEPASammler.create()
        context.commit()
        super(SEPASammlerAddView, self).__init__(context, request)
        self.log()

    def log(self):
        log_with_user(
            user_data_log.info, self.request.user,
            'SEPA Sammler %s hinzugefügt.', self.context.id)

    @property
    def route_name(self):
        return 'sepa_sammler_edit'


@view_config(route_name='sepa_ueberweisung_add', renderer='json',
             permission='view')
class SEPAUeberweisungAddView(SEPASammlerAddView):

    def log(self):
        self.context.is_ueberweisung = True
        log_with_user(
            user_data_log.info, self.request.user,
            'SEPA Überweisung %s hinzugefügt.', self.context.id)


@view_config(route_name='sepa_sammler_export', renderer='json',
             permission='view')
class SEPASammlerExportView(sw.allotmentclub.browser.base.SEPAExporterView):

    @property
    def filename(self):
        return '{} {}'.format(
            self.context.kind.title, self.context.accounting_year)

    @property
    def subject(self):
        return self.filename

    @property
    def values(self):
        return sorted(self.context.entries, key=lambda x: x.member.lastname)

    def get_member(self, value):
        return value.member

    def get_to_pay(self, value):
        return value.value

    @property
    def creation_date(self):
        return datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')

    @property
    def booking_day(self):
        return self.context.booking_day.strftime('%Y-%m-%d')


class SEPASammlerEntryQuery(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Betrag': format_eur,
        'Ignorieren': boolean,
    }

    data_class = {
        'Zugeordnet': 'expand'
    }

    data_hide = {
        'Ignorieren': 'phone,tablet',
        'Betrag': 'phone'
    }

    def select(self):
        return (
            self.db.query(
                SEPASammlerEntry.id.label('#'),
                (to_string(Member.lastname).concat(', ').concat(
                 to_string(Member.firstname).concat(' (')
                 .concat(func.coalesce(string_agg(Allotment.number), 'n/a'))
                 .concat(')'))).label('Zugeordnet'),
                SEPASammlerEntry.value.label('Betrag'),
                SEPASammlerEntry.ignore_in_reporting.label('Ignorieren')
            )
            .select_from(SEPASammlerEntry)
            .filter(SEPASammlerEntry.sepasammler == self.context)
            .join(Member)
            .outerjoin(Allotment, Allotment.member_id == Member.id)
            .group_by(SEPASammlerEntry.id, Member.lastname, Member.firstname))


@view_config(route_name='sepa_sammler_entry_list', renderer='json',
             permission='view')
class SEPASammlerEntryListView(sw.allotmentclub.browser.base.TableView):

    query_class = SEPASammlerEntryQuery
    year_selection = False


def account_holder(id, request=None):
    member = Member.get(id)
    result = member.direct_debit_account_holder or (
        '%s, %s' % (member.lastname, member.firstname))
    return '%s (%s)' % (
        result, '/'.join(str(a.number) for a in member.allotments))


class SEPADirectDebitQuery(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Datum': date,
        'Mitglied': account_holder,
    }
    data_class = {
        'Mitglied': 'expand'
    }
    data_hide = {
        'IBAN': 'phone',
        'BIC': 'phone,tablet',
        'Datum': 'phone,tablet',
    }

    def select(self):
        from sqlalchemy.sql.expression import true
        return (
            self.db.query(
                Member.id.label('#'),
                Member.id.label('Mitglied'),
                Member.iban.label('IBAN'),
                Member.bic.label('BIC'),
                Member.direct_debit_date.label('Datum'),
            )
            .select_from(Member)
            .join(Allotment, Allotment.member_id == Member.id)
            .group_by(Member.id)
            .filter(Member.direct_debit == true())
        )


@view_config(route_name='sepa_direct_debit', renderer='json',
             permission='view')
class SEPADirectDebitListView(sw.allotmentclub.browser.base.TableView):

    query_class = SEPADirectDebitQuery



# encoding=utf8
"""re #2: Add change of ElectricMeter 14996103.

Revision ID: 444966e3e7d6
Revises: 56e61da4fed1
Create Date: 2014-12-21 17:10:21.186765

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '444966e3e7d6'
down_revision = '56e61da4fed1'


def upgrade():
    op.add_column(
        'energyvalue',
        sa.Column('member_id', sa.Integer(), nullable=True))


def downgrade():
    op.drop_column('energyvalue', 'member_id')



"""Add access authority

Revision ID: a08b46e0172e
Revises: 4b98c95b9a56
Create Date: 2016-07-11 15:35:47.894579

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'a08b46e0172e'
down_revision = '4b98c95b9a56'

LEVELS = {
    'portal-revision': (5, 4, 2),
    'portal-club-admin': (2,),
    'portal-read-only': (4, 2),
    'portal-admin': (),
}

AUTHORITIES = (
    ('booking_list', 'portal-revision'),
    ('split_booking', 'portal-club-admin'),
    ('map_booking', 'portal-club-admin'),
    ('member_account_list', 'portal-revision'),
    ('member_account_detail_list', 'portal-revision'),
    ('member_account_detail_switch_ir', 'portal-club-admin'),
    ('banking_account_list', 'portal-revision'),
    ('sepa_sammler_list', 'portal-club-admin'),
    ('sepa_sammler_update', 'portal-read-only'),
    ('sepa_sammler_edit', 'portal-read-only'),
    ('sepa_sammler_add', 'portal-read-only'),
    ('sepa_sammler_export', 'portal-read-only'),
    ('sepa_sammler_entry_list', 'portal-club-admin'),
    ('sepa_direct_debit', 'portal-club-admin'),
    ('assignments', 'portal-read-only'),
    ('assignment_edit', 'portal-read-only'),
    ('assignment_add', 'portal-read-only'),
    ('assignment_delete', 'portal-admin'),
    ('assignment_list_attendees', 'portal-read-only'),
    ('assignment_attendees_edit', 'portal-read-only'),
    ('assignment_attendees_add', 'portal-read-only'),
    ('assignment_attendees_delete', 'portal-read-only'),
    ('member_assignments', 'portal-read-only'),
    ('member_assignments_bill', 'portal-admin'),
    ('access_authority', 'portal-admin'),
    ('access_authority_detail', 'portal-admin'),
    ('access_authority_detail_edit', 'portal-admin'),
    ('access_authority_detail_add', 'portal-admin'),
    ('access_authority_detail_delete', 'portal-admin'),
    ('bulletins', 'portal-read-only'),
    ('bulletin_edit', 'portal-read-only'),
    ('bulletin_add', 'portal-read-only'),
    ('bulletin_delete', 'portal-admin'),
    ('bulletin_print', 'portal-read-only'),
    ('depots', 'portal-read-only'),
    ('depot_add', 'portal-read-only'),
    ('depot_edit', 'portal-club-admin'),
    ('depot_delete', 'portal-admin'),
    ('depot_download', 'portal-read-only'),
    ('electricity_list', 'portal-revision'),
    ('global_energy_value_list', 'portal-revision'),
    ('energy_value_list', 'portal-revision'),
    ('advance_pay_value_list', 'portal-read-only'),
    ('energy_meter_export', 'portal-read-only'),
    ('energy_meter_import', 'portal-admin'),
    ('externals', 'portal-read-only'),
    ('external_edit', 'portal-read-only'),
    ('external_add', 'portal-read-only'),
    ('home', 'portal-read-only'),
    ('keylists', 'portal-read-only'),
    ('keylist_edit', 'portal-read-only'),
    ('keylist_add', 'portal-read-only'),
    ('keylist_delete', 'portal-admin'),
    ('keys', 'portal-read-only'),
    ('key_edit', 'portal-read-only'),
    ('key_add', 'portal-read-only'),
    ('key_delete', 'portal-admin'),
    ('mail_list', 'portal-read-only'),
    ('mail_preview', 'portal-read-only'),
    ('mail_print', 'portal-read-only'),
    ('mail_send', 'portal-read-only'),
    ('mail_duplicate', 'portal-read-only'),
    ('mail_delete', 'portal-read-only'),
    ('mail_edit', 'portal-read-only'),
    ('mail_add', 'portal-read-only'),
    ('mail_list_attachments', 'portal-read-only'),
    ('mail_attachment_white_page', 'portal-read-only'),
    ('mail_attachment_del', 'portal-read-only'),
    ('mail_attachment_download', 'portal-read-only'),
    ('map', 'portal-read-only'),
    ('map_download', 'portal-read-only'),
    ('member_list', 'portal-read-only'),
    ('member_edit', 'portal-club-admin'),
    ('member_address_verify', 'portal-admin'),
    ('direct_debit_letter', 'portal-read-only'),
    ('become_member_letter', 'portal-read-only'),
    ('member_sale', 'portal-admin'),
    ('mv_entrance_list', 'portal-read-only'),
    ('member_sale_history', 'portal-read-only'),
    ('protocols', 'portal-read-only'),
    ('protocol_edit', 'portal-admin'),
    ('protocol_add', 'portal-admin'),
    ('protocol_delete', 'portal-admin'),
    ('protocol_detail', 'portal-read-only'),
    ('protocol_detail_edit', 'portal-admin'),
    ('protocol_detail_add', 'portal-admin'),
    ('protocol_detail_delete', 'portal-admin'),
    ('protocol_attachment', 'portal-read-only'),
    ('protocol_attachment_add', 'portal-admin'),
    ('protocol_attachment_download', 'portal-read-only'),
    ('protocol_commitment', 'portal-read-only'),
    ('protocol_commitment_edit', 'portal-admin'),
    ('protocol_commitment_add', 'portal-admin'),
    ('protocol_commitment_delete', 'portal-admin'),
    ('protocol_print', 'portal-read-only'),
)


def upgrade():
    op.create_table(
        'accessauthority',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('viewname', sa.String(), nullable=True),
        sa.Column('user_id', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(['user_id'], [u'user.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    for viewname, group in AUTHORITIES:
        for user_id in LEVELS[group]:
            op.execute("""INSERT INTO accessauthority (viewname, user_id)
                          VALUES ('{}', {});""".format(viewname, user_id))


def downgrade():
    op.drop_table('accessauthority')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from .model import Object
from .base import format_kwh, format_eur
from sqlalchemy import Column, Integer, String, ForeignKey, Boolean, sql
from sw.allotmentclub.account import Booking, BookingKind, BankingAccount
import datetime
import sqlalchemy.orm

THRESHOLD = 100000.0


class ElectricMeter(Object):
    """ Der Stromzähler."""

    allotment_id = Column(Integer, ForeignKey('allotment.id'), nullable=False)
    allotment = sqlalchemy.orm.relation(
        'Allotment', uselist=False, backref='electric_meters')
    energy_values = sqlalchemy.orm.relation(
        'EnergyValue', uselist=True, backref='electric_meter',
        cascade="delete")
    number = Column(String)
    electric_power = Column(Boolean, default=sql.false())
    disconnected = Column(Boolean, default=sql.false())
    comment = Column(String)
    replaced_by_id = Column(
        Integer, ForeignKey('electricmeter.id'), nullable=True)
    replaced_by = sqlalchemy.orm.relation('ElectricMeter', uselist=False)
    discount_to_id = Column(
        Integer, ForeignKey('member.id'), nullable=True)
    discount_to = sqlalchemy.orm.relation('Member', uselist=False)

    @classmethod
    def by_number(cls, number):
        return cls.query().filter(cls.number == str(number)).one_or_none()

    def get_value(self, year):
        for value in self.energy_values:
            if value.year == year:
                return value

    def get_last_value(self, exclude=None):
        if (len(self.energy_values) == 1):
            return self.energy_values[0]
        start = min([v.year for v in self.energy_values])
        last_value = self.get_value(start)
        for value in self.energy_values:
            if value is exclude:
                continue
            if ((getattr(value, 'value', 0) or 0) >
                    (getattr(last_value, 'value', 0) or 0)):
                last_value = value
        return last_value

    def replace(self, old_value, new_number, new_value):
        year = datetime.datetime.now().year
        new = ElectricMeter()
        new.allotment = self.allotment
        new.number = new_number
        new.electric_power = self.electric_power
        self.replaced_by = new
        new.discount_to = self.discount_to
        new.organization_id = self.organization_id

        last_value = self.get_last_value()
        if last_value.value < old_value:
            value = EnergyValue.create(
                electric_meter=self, year=year, value=old_value,
                organization_id=self.organization_id)
            value.update_member()
            value.update_usage()
            self.energy_values.append(value)

        value = EnergyValue.create(
            electric_meter=new, year=year-1, value=new_value,
            organization_id=self.organization_id)
        value.update_member()
        value.advance_pay = value.fee = value.price = value.usage = 0
        value.whole_price = value.to_pay = 0
        value.discounted = True
        new.energy_values.append(value)
        self.disconnected = True
        return new


class EnergyValue(Object):
    """ Der Zählerstand."""

    electric_meter_id = Column(Integer, ForeignKey(ElectricMeter.id),
                               nullable=False)
    year = Column(Integer)
    value = Column(Integer)
    estimated_value = Column(Boolean, default=sql.false())
    member_id = Column(Integer, ForeignKey('member.id'), nullable=True)
    member = sqlalchemy.orm.relation(
        'Member', uselist=False, backref='energy_values')
    usage = Column(Integer, nullable=True)
    price = Column(Integer, nullable=True)
    fee = Column(Integer, nullable=True)
    whole_price = Column(Integer, nullable=True)
    to_pay = Column(Integer, nullable=True)
    advance_pay = Column(Integer, nullable=True)
    discounted = Column(Boolean, nullable=True)

    def update_member(self):
        if not self.member:
            if self.electric_meter.discount_to:
                self.member = self.electric_meter.discount_to
            else:
                self.member = self.electric_meter.allotment.member

    def update_usage(self):
        self.usage = self._usage

    def update_data(self):
        from sw.allotmentclub.browser.account import value_to_int
        from sw.allotmentclub.browser.base import format_eur
        if self.discounted:
            return
        self.price = self._price
        self.fee = self._fee
        self.whole_price = self._whole_price
        self.to_pay = self._to_pay
        if self._advance_pay >= THRESHOLD:
            self.advance_pay = int(self._advance_pay)
        else:
            self.advance_pay = 0
        purpose = 'Energieabrechnung'
        month = 8
        day = 31
        try:
            kind = (BookingKind.query()
                    .filter(BookingKind.title == purpose).one())
        except sqlalchemy.orm.exc.NoResultFound:
            pass
        else:
            account = (BankingAccount.query()
                       .filter(BankingAccount.number == '3').one())
            booking = Booking.find_or_create(
                banking_account=account,
                purpose='{} für Zähler {}'.format(
                    purpose, self.electric_meter.number),
                accounting_year=self.year,
                booking_day=datetime.date(self.year, month, day),
                member=self.member,
                kind=kind)
            booking.value = 0 - value_to_int(format_eur(self.to_pay)[:-2])
        if self.member.lastname in ['Verein', 'Werkstatt']:
            return
        for purpose, month, day in [('Energieabschlag I', 3, 31),
                                    ('Energieabschlag II', 6, 30)]:
            try:
                kind = (BookingKind.query()
                        .filter(BookingKind.title == purpose).one())
            except sqlalchemy.orm.exc.NoResultFound:
                continue
            account = (BankingAccount.query()
                       .filter(BankingAccount.number == '3').one())
            booking = Booking.find_or_create(
                banking_account=account,
                purpose='{} für Zähler {}'.format(
                    purpose, self.electric_meter.number),
                accounting_year=self.year+1,
                booking_day=datetime.date(self.year+1, month, day),
                member=self.member,
                kind=kind)
            booking.value = 0 - value_to_int(format_eur(self.advance_pay)[:-2])

    @property
    def _usage(self):
        before = self.electric_meter.get_last_value(self)
        if before is None or self.value is None or before.value is None:
            return 0
        return self.value - before.value

    @property
    def _price(self):
        try:
            price = EnergyPrice.query().filter(
                EnergyPrice.year == self.year).one()
        except Exception:
            return 0
        if not price.price:
            return 0
        return self.usage * price.price

    @property
    def _fee(self):
        try:
            price = EnergyPrice.query().filter(
                EnergyPrice.year == self.year).one()
        except Exception:
            return 0
        if self.electric_meter.disconnected:
            return 0
        if not price.power_fee or not price.normal_fee:
            return 0
        if self.electric_meter.electric_power:
            return price.power_fee
        return price.normal_fee

    @property
    def _whole_price(self):
        return self.fee + self.price

    @property
    def advance_pay_last_year(self):
        last_value = self.electric_meter.get_value(self.year-1)
        if last_value:
            advance_pay_last_year = last_value.advance_pay
            advance_pay_last_year = int(
                round(advance_pay_last_year / 10000.0, 2) * 10000
            )
        else:
            advance_pay_last_year = 0
        return advance_pay_last_year * 2

    @property
    def _to_pay(self):
        return self.whole_price - self.advance_pay_last_year

    @property
    def _advance_pay(self):
        # Calculate the advance payment for the next year
        return self.whole_price / 3.0


class EnergyPrice(Object):
    """ Der Preis einer kWh. """

    year = Column(Integer)
    value = Column(Integer)  # Stand Hauptzähler zur Abrechnung
    bill = Column(Integer)  # Abrechnung Energieanbieter in Eur
    usage_hauptzaehler = Column(Integer)  # Verbrauch Hauptzaehler
    usage_members = Column(Integer)  # Verbrauch Mitglieder
    leakage_current = Column(Integer)  # Verluststrom
    price = Column(Integer)  # Preis in MicroCent pro kWh
    normal_fee = Column(Integer)  # Gebühr normaler Zähler in MicroCent
    power_fee = Column(Integer)  # Gebühr Kraftstrom-Zähler in MicroCent


def format_iban(iban):
    if not iban or len(iban) < 10:
        return iban
    return 'X' * 10 + iban[10:]


def get_energyvalue_mail_data(member, year):
    for value in (EnergyValue.query()
                  .filter(EnergyValue.year == year)
                  .filter(EnergyValue.member == member)).all():
        price = (
            EnergyPrice.query().filter(EnergyPrice.year == value.year).one())
        under_threshold = not value.advance_pay
        above_threshold = value.advance_pay
        if value.electric_meter.disconnected:
            under_threshold = False
            above_threshold = False
        content_data = dict(
            deflection='' if member.appellation == 'Frau' else 'r',
            appellation=member.appellation,
            title=member.title,
            lastname=member.direct_debit_account_holder or member.lastname,
            value=format_kwh(value.value),
            usage=format_kwh(value.usage),
            value_last_year=format_kwh(value.value-value.usage),
            meter=value.electric_meter.number,
            whole_price=format_eur(value.whole_price),
            to_pay=format_eur(value.to_pay),
            last_year=value.year-1,
            year=value.year,
            next_year=value.year+1,
            iban=format_iban(member.iban),
            bic=member.bic,
            threshold=format_eur(THRESHOLD),
            advance_pay=format_eur(value.advance_pay_last_year),
            advance_pay_next_year=format_eur(value._advance_pay),
            price_kwh=format_eur(price.price, full=True),
            normal_fee=format_eur(price.normal_fee),
            power_fee=format_eur(price.power_fee),
            no_abschlag=value.to_pay == value.whole_price,
            must_pay=value.to_pay > 0,
            gets_back=value.to_pay <= 0,
            disconnected=value.electric_meter.disconnected,
            under_threshold=under_threshold,
            above_threshold=above_threshold)
        subject = u'Energieabrechnung für Zähler Nr. %s' % (
            value.electric_meter.number)
        yield subject, content_data



"""Add attachments to parcels

Revision ID: 9af3776b0ea9
Revises: ecca91384800
Create Date: 2019-11-05 08:28:17.231686

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '9af3776b0ea9'
down_revision = 'ecca91384800'


def upgrade():
    op.create_table(
        'parcelattachment',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('organization_id', sa.Integer(), nullable=False),
        sa.Column('parcel_id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(length=100), nullable=True),
        sa.Column('mimetype', sa.String(length=100), nullable=True),
        sa.Column('size', sa.String(length=20), nullable=True),
        sa.Column('data', sa.LargeBinary(length=10485760), nullable=True),
        sa.ForeignKeyConstraint(['parcel_id'], ['parcel.id'], ),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('parcelattachment')



"""Add protocol details.

Revision ID: 571eb56a2b04
Revises: 4ac1a4b79213
Create Date: 2014-12-16 22:08:56.474723

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '571eb56a2b04'
down_revision = '4ac1a4b79213'


def upgrade():
    op.create_table(
        'protocoldetail',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('protocol_id', sa.Integer(), nullable=False),
        sa.Column('duration', sa.Integer(), nullable=True),
        sa.Column('message', sa.Text(), nullable=True),
        sa.Column('responsible', sa.String(length=2), nullable=True),
        sa.ForeignKeyConstraint(['protocol_id'], [u'protocol.id'], ),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('protocoldetail')



# encoding=utf-8
import sw.allotmentclub.version


def test_no_access_if_not_logged_in(browser):
    browser.open('http://localhost/login')
    assert {
        u'message': u'',
        u'status': u'error',
        u'version': sw.allotmentclub.version.__version__
    } == browser.json


def test_login_is_possible(browser):
    browser.login()
    result = browser.json
    assert 'Anmeldung erfolgreich.' == result['message']
    assert 'success' == result['status']
    assert sw.allotmentclub.version.__version__ == result['version']
    assert {
        u'gravatar': (u'https://www.gravatar.com/avatar/'
                      u'd41d8cd98f00b204e9800998ecf8427e'),
        u'name': u'Admin istrator',
        u'username': u'admin'
    } == result['user']


def test_no_access_if_logging_in_with_invalid_data(browser):
    browser.login(username='admin', password='wrong')
    assert {
        u'message': u'Anmeldung fehlgeschlagen.',
        u'status': u'error',
        u'version': sw.allotmentclub.version.__version__
    } == browser.json


def test_double_login_not_possible(browser):
    browser.login()
    browser.login()
    assert u'Ihre Sitzung wurde wiederhergestellt.' == browser.json['message']


def test_logout_is_possible(browser):
    browser.login()
    browser.open('http://localhost/logout')
    assert {
        u'message': (u'Abmeldung abgeschlossen. Sie können sich nun '
                     u'erneut anmelden.'),
        u'status': u'success',
        u'version': sw.allotmentclub.version.__version__
    } == browser.json



"""Add sale history

Revision ID: c3c978348cbf
Revises: cc8ba97635c1
Create Date: 2016-02-03 21:14:16.233171

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'c3c978348cbf'
down_revision = 'cc8ba97635c1'


def upgrade():
    op.create_table(
        'salehistory',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('seller_id', sa.Integer(), nullable=False),
        sa.Column('buyer_id', sa.Integer(), nullable=False),
        sa.Column('date', sa.Date(), nullable=False),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('salehistory')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import datetime
import transaction


def setUp():
    from sw.allotmentclub import Depot, User
    user = User.create(username='hans')
    Depot.create(date=datetime.datetime(2014, 11, 27, 7, 21, 45),
                 size=15, data=b'GIF89a????!?,D;', mimetype='image/gif',
                 name='test.gif', user=user)
    transaction.commit()


def test_displays_list_of_depots(browser):
    setUp()
    browser.login()
    browser.open('http://localhost/depots')
    expected = [
        1, 'test.gif', '', 'GIF', '15.00 B', '27.11.2014 07:21', 'hans'
    ]
    assert expected in browser.json_result


def test_depot_can_be_added_via_json_view(browser):
    from sw.allotmentclub import Depot
    browser.login()
    browser._upload('http://localhost/depots/add')
    assert 'success' == browser.json['status']
    assert 'Transparent.gif' == Depot.query().one().name


def test_depot_can_be_edited_via_json_view(browser, json_fixture):
    from sw.allotmentclub import Depot
    setUp()
    url = json_fixture.url()
    browser.login()
    browser.post(
        'http://localhost{}'.format(url),
        data={"name": "foo.gif"}
    )
    assert 'success' == browser.json['status']
    assert Depot.get(1).name == 'foo.gif'

    browser.post(
        'http://localhost{}'.format(url),
        data={"category": "Mitgliederversammlung"}
    )
    assert 'success' == browser.json['status']
    assert Depot.get(1).category == 'Mitgliederversammlung'



"""Add signature for st

Revision ID: 38a1521607b5
Revises: 2fd64f1e524c
Create Date: 2015-01-04 15:40:04.177104

"""
from alembic import op
import sqlalchemy as sa
import base64
import pkg_resources

# revision identifiers, used by Alembic.
revision = '38a1521607b5'
down_revision = '2fd64f1e524c'


def upgrade():
    op.drop_column('user', 'signature')
    op.add_column(
        'user',
        sa.Column('signature', sa.String(), nullable=True))
    signature = pkg_resources.resource_stream(
        'sw.allotmentclub.signatures', 'st.png').read()
    signature = 'data:application/png;base64,{}'.format(
        base64.b64encode(signature))
    op.execute(u"""UPDATE public.user SET
                  signature = '{signature}'
                  WHERE username = 'st';""".format(signature=signature))


def downgrade():
    op.drop_column('user', 'signature')
    op.add_column(
        'user',
        sa.Column('signature', sa.LargeBinary(length=10485760), nullable=True))
